import functools
from collections import OrderedDict

import torch
from mmseg.models import SEGMENTORS, build_segmentor
from mmseg.models.segmentors.base import BaseSegmentor
from mmseg.ops import resize

from otx.mpa.utils.logger import get_logger

logger = get_logger()


@SEGMENTORS.register_module()
class MeanTeacherSegmentor(BaseSegmentor):
    def __init__(self, orig_type=None, unsup_weight=0.1, semisl_start_iter=30, **kwargs):
        super(MeanTeacherSegmentor, self).__init__()
        self.test_cfg = kwargs["test_cfg"]
        self.semisl_start_iter = semisl_start_iter
        self.count_iter = 0

        cfg = kwargs.copy()
        cfg["type"] = orig_type
        self.align_corners = cfg["decode_head"].align_corners
        self.model_s = build_segmentor(cfg)
        self.model_t = build_segmentor(cfg)
        self.unsup_weight = unsup_weight

        # Hooks for super_type transparent weight load/save
        self._register_state_dict_hook(self.state_dict_hook)
        self._register_load_state_dict_pre_hook(functools.partial(self.load_state_dict_pre_hook, self))

    def encode_decode(self, img, img_metas):
        return self.model_s.encode_decode(img, img_metas)

    def extract_feat(self, imgs):
        return self.model_s.extract_feat(imgs)

    def simple_test(self, img, img_metas, **kwargs):
        return self.model_s.simple_test(img, img_metas, **kwargs)

    def aug_test(self, imgs, img_metas, **kwargs):
        return self.model_s.aug_test(imgs, img_metas, **kwargs)

    def forward_dummy(self, img, **kwargs):
        return self.model_s.forward_dummy(img, **kwargs)

    def forward_train(self, img, img_metas, gt_semantic_seg, **kwargs):
        self.count_iter += 1
        if self.semisl_start_iter > self.count_iter or "extra_0" not in kwargs:
            x = self.model_s.extract_feat(img)
            loss_decode = self.model_s._decode_head_forward_train(x, img_metas, gt_semantic_seg=gt_semantic_seg)
            return loss_decode

        ul_data = kwargs["extra_0"]
        ul_s_img = ul_data["img"]
        ul_w_img = ul_data["ul_w_img"]
        ul_img_metas = ul_data["img_metas"]

        with torch.no_grad():
            teacher_feat = self.model_t.extract_feat(ul_w_img)
            teacher_logit = self.model_t._decode_head_forward_test(teacher_feat, ul_img_metas)
            teacher_logit = resize(
                input=teacher_logit, size=ul_w_img.shape[2:], mode="bilinear", align_corners=self.align_corners
            )
            conf_from_teacher, pl_from_teacher = torch.max(torch.softmax(teacher_logit, axis=1), axis=1, keepdim=True)

        losses = dict()

        x = self.model_s.extract_feat(img)
        x_u = self.model_s.extract_feat(ul_s_img)
        loss_decode = self.model_s._decode_head_forward_train(x, img_metas, gt_semantic_seg=gt_semantic_seg)
        loss_decode_u = self.model_s._decode_head_forward_train(x_u, ul_img_metas, gt_semantic_seg=pl_from_teacher)

        for (k, v) in loss_decode_u.items():
            if v is None:
                continue
            losses[k] = loss_decode[k] + loss_decode_u[k] * self.unsup_weight

        return losses

    @staticmethod
    def state_dict_hook(module, state_dict, prefix, *args, **kwargs):
        """Redirect student model as output state_dict (teacher as auxilliary)"""
        logger.info("----------------- MeanTeacherSegmentor.state_dict_hook() called")
        for k in list(state_dict.keys()):
            v = state_dict.pop(k)
            if not prefix or k.startswith(prefix):
                k = k.replace(prefix, "", 1)
                if k.startswith("model_s."):
                    k = k.replace("model_s.", "", 1)
                elif k.startswith("model_t."):
                    continue
                k = prefix + k
            state_dict[k] = v
        return state_dict

    @staticmethod
    def load_state_dict_pre_hook(module, state_dict, *args, **kwargs):
        """Redirect input state_dict to teacher model"""
        logger.info("----------------- MeanTeacherSegmentor.load_state_dict_pre_hook() called")
        for k in list(state_dict.keys()):
            v = state_dict.pop(k)
            state_dict["model_s." + k] = v
            state_dict["model_t." + k] = v



"""This module define the annotation entity."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc
import datetime
from enum import Enum
from typing import Dict, List, Optional, Set

from bson import ObjectId

from otx.api.entities.id import ID
from otx.api.entities.label import LabelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.shape import ShapeEntity
from otx.api.utils.time_utils import now


class Annotation(metaclass=abc.ABCMeta):
    """Base class for annotation objects.

    Args:
        shape (ShapeEntity): the shape of the annotation
        labels (List[ScoredLabel]): the labels of the annotation
        id (Optional[ID]): the id of the annotation
    """

    # pylint: disable=redefined-builtin;
    def __init__(self, shape: ShapeEntity, labels: List[ScoredLabel], id: Optional[ID] = None):
        self.__id_ = ID(ObjectId()) if id is None else id
        self.__shape = shape
        self.__labels = labels

    def __repr__(self):
        """String representation of the annotation."""
        return (
            f"{self.__class__.__name__}("
            f"shape={self.shape}, "
            f"labels={self.get_labels(include_empty=True)}, "
            f"id={self.id_})"
        )

    @property
    def id_(self):
        """Returns the id for the annotation."""
        return self.__id_

    @id_.setter
    def id_(self, value):
        self.__id_ = value

    @property
    def id(self):
        """DEPRECATED."""
        return self.__id_

    @id.setter
    def id(self, value):
        """DEPRECATED."""
        self.__id_ = value

    @property
    def shape(self) -> ShapeEntity:
        """Returns the shape that is in the annotation."""
        return self.__shape

    @shape.setter
    def shape(self, value) -> None:
        self.__shape = value

    def get_labels(self, include_empty: bool = False) -> List[ScoredLabel]:
        """Get scored labels that are assigned to this annotation.

        Args:
            include_empty (bool): set to True to include empty label (if exists) in the output. Defaults to False.

        Returns:
            List of labels in annotation
        """
        return [label for label in self.__labels if include_empty or (not label.is_empty)]

    def get_label_ids(self, include_empty: bool = False) -> Set[ID]:
        """Get a set of ID's of labels that are assigned to this annotation.

        Args:
            include_empty (bool): set to True to include empty label (if exists) in the output. Defaults to False.

        Returns:
            Set of label id's in annotation
        """
        return {label.id_ for label in self.__labels if include_empty or (not label.is_empty)}

    def append_label(self, label: ScoredLabel) -> None:
        """Appends the scored label to the annotation.

        Args:
            label (ScoredLabel): the scored label to be appended to the annotation
        """
        self.__labels.append(label)

    def set_labels(self, labels: List[ScoredLabel]) -> None:
        """Sets the labels of the annotation to be the input of the function.

        Args:
            labels (List[ScoredLabel]): the scored labels to be set as annotation labels
        """
        self.__labels = labels

    def __eq__(self, other: object) -> bool:
        """Checks if the two annotations are equal.

        Args:
            other (Annotation): Annotation to compare with.

        Returns:
            bool: True if the two annotations are equal, False otherwise.
        """
        if isinstance(other, Annotation):
            return (
                self.id_ == other.id_ and self.get_labels(True) == other.get_labels(True) and self.shape == other.shape
            )
        return False


class AnnotationSceneKind(Enum):
    """AnnotationSceneKinds for an Annotation object."""

    #:  NONE represents NULLAnnotationScene's (See :class:`NullAnnotationScene`)
    NONE = 0
    #:  ANNOTATION represents user annotation
    ANNOTATION = 1
    #:  PREDICTION represents analysis result, which will be shown to the user
    PREDICTION = 2
    #:  EVALUATION represents analysis result for evaluation purposes, which will NOT be shown to the user
    EVALUATION = 3
    #:  INTERMEDIATE represents intermediary state.
    #:  This is used when the analysis is being transferred from one task to another.
    #:  This will not be shown to the user.
    #:  This state will be changed to either PREDICTION or EVALUATION at the end of analysis process.
    INTERMEDIATE = 4
    #:  TASK_PREDICTION represents analysis results for a single task
    TASK_PREDICTION = 5

    def __str__(self):
        """String representation of the AnnotationSceneKind."""
        return str(self.name)


class AnnotationSceneEntity(metaclass=abc.ABCMeta):
    """This class represents a user annotation or a result (prediction).

    It serves as a collection of shapes, with a relation to the media entity.

    Example:
        Creating an annotation:

        >>> from otx.api.entities.annotation import Annotation, AnnotationSceneEntity, AnnotationSceneKind
        >>> from otx.api.entities.shapes.rectangle import Rectangle
        >>> box = Rectangle(x1=0.0, y1=0.0, x2=0.5, y2=0.5)  # Box covering top-left quart of image
        >>> AnnotationSceneEntity(annotations=[Annotation(shape=box, labels=[])], kind=AnnotationSceneKind.ANNOTATION)

    Args:
        annotations (List[Annotation]): List of annotations in the scene
        kind (AnnotationSceneKind): Kind of the annotation scene. E.g. `AnnotationSceneKind.ANNOTATION`.
        editor (str): The user that made this annotation scene object.
        creation_date (Optional[datetime.datetime]): Creation date of annotation scene entity. If None, current time is
            used. Defaults to None.
        id (Optional[ID]): ID of AnnotationSceneEntity. If None a new `ID` is created. Defaults to None.
    """

    # pylint: disable=too-many-arguments, redefined-builtin
    def __init__(
        self,
        annotations: List[Annotation],
        kind: AnnotationSceneKind,
        editor: str = "",
        creation_date: Optional[datetime.datetime] = None,
        id: Optional[ID] = None,
    ):
        self.__annotations = annotations
        self.__kind = kind
        self.__editor = editor
        self.__creation_date = now() if creation_date is None else creation_date
        self.__id_ = ID() if id is None else id

    def __repr__(self):
        """String representation of the annotation scene."""
        return (
            f"{self.__class__.__name__}("
            f"annotations={self.annotations}, "
            f"kind={self.kind}, "
            f"editor={self.editor_name}, "
            f"creation_date={self.creation_date}, "
            f"id={self.id_})"
        )

    @property
    def id_(self) -> ID:
        """Returns the ID of the AnnotationSceneEntity."""
        return self.__id_

    @id_.setter
    def id_(self, value) -> None:
        self.__id_ = value

    @property
    def id(self):
        """DEPRECATED."""
        return self.__id_

    @id.setter
    def id(self, value):
        """DEPRECATED."""
        self.__id_ = value

    @property
    def kind(self) -> AnnotationSceneKind:
        """Returns the AnnotationSceneKind of the AnnotationSceneEntity."""
        return self.__kind

    @kind.setter
    def kind(self, value) -> None:
        self.__kind = value

    @property
    def editor_name(self) -> str:
        """Returns the editor's name that made the AnnotationSceneEntity object."""
        return self.__editor

    @editor_name.setter
    def editor_name(self, value) -> None:
        self.__editor = value

    @property
    def creation_date(self) -> datetime.datetime:
        """Returns the creation date of the AnnotationSceneEntity object."""
        return self.__creation_date

    @creation_date.setter
    def creation_date(self, value) -> None:
        self.__creation_date = value

    @property
    def annotations(self) -> List[Annotation]:
        """Return the Annotations that are present in the AnnotationSceneEntity."""
        return self.__annotations

    @annotations.setter
    def annotations(self, value: List[Annotation]):
        self.__annotations = value

    @property
    def shapes(self) -> List[ShapeEntity]:
        """Returns all shapes that are inside the annotations of the AnnotationSceneEntity."""
        return [annotation.shape for annotation in self.annotations]

    def contains_any(self, labels: List[LabelEntity]) -> bool:
        """Checks whether the annotation contains any labels in the input parameter.

        Args:
            labels (List[LabelEntity]): List of labels to compare to.

        Returns:
            bool: True if there is any intersection between self.get_labels(include_empty=True) with labels.
        """
        label_names = {label.name for label in labels}
        return len({label.name for label in self.get_labels(include_empty=True)}.intersection(label_names)) != 0

    def append_annotation(self, annotation: Annotation) -> None:
        """Appends the passed annotation to the list of annotations present in the AnnotationSceneEntity object."""
        self.annotations.append(annotation)

    def append_annotations(self, annotations: List[Annotation]) -> None:
        """Adds a list of annotations to the annotation scene."""
        self.annotations.extend(annotations)

    def get_labels(self, include_empty: bool = False) -> List[LabelEntity]:
        """Returns a list of unique labels which appear in this annotation scene.

        Args:
            include_empty (bool): Set to True to include empty label (if exists) in the output. Defaults to False.

        Returns:
            List[LabelEntity]: a list of labels which appear in this annotation.
        """

        labels: Dict[str, LabelEntity] = {}
        for annotation in self.annotations:
            for label in annotation.get_labels(include_empty=include_empty):
                id_ = label.id_
                if id_ not in labels:
                    labels[id_] = label.get_label()
        return list(labels.values())

    def get_label_ids(self, include_empty: bool = False) -> Set[ID]:
        """Returns a set of the ID's of unique labels which appear in this annotation scene.

        Args:
            include_empty (bool): Set to True to include empty label (if exists) in the output. Defaults to False.

        Returns:
            Set[ID]: a set of the ID's of labels which appear in this annotation.
        """

        output: Set[ID] = set()
        for annotation in self.annotations:
            output.update(set(annotation.get_label_ids(include_empty=include_empty)))
        return output


class NullAnnotationSceneEntity(AnnotationSceneEntity):
    """Represents 'AnnotationSceneEntity not found."""

    def __init__(self) -> None:
        super().__init__(
            id=ID(),
            kind=AnnotationSceneKind.NONE,
            editor="",
            creation_date=datetime.datetime.now(),
            annotations=[],
        )

    def __repr__(self):
        """String representation NullAnnotationSceneEntity."""
        return "NullAnnotationSceneEntity()"



"""Dataset utils."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import List, Optional, Tuple

import numpy as np

from otx.api.entities.annotation import AnnotationSceneEntity
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import LabelEntity
from otx.api.entities.model import ModelEntity
from otx.api.entities.result_media import ResultMediaEntity
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.utils.vis_utils import get_actmap


def get_fully_annotated_idx(dataset: DatasetEntity) -> List[int]:
    """Find the indices of the fully annotated items in a dataset.

    A dataset item is fully annotated if local annotations are available, or if the item has the `normal` label.

    Args:
        dataset (DatasetEntity): Dataset that may contain both partially and fully annotated items.

    Returns:
        List[int]: List of indices of the fully annotated dataset items.
    """
    local_idx = []
    for idx, gt_item in enumerate(dataset):
        local_annotations = [
            annotation for annotation in gt_item.get_annotations() if not Rectangle.is_full_box(annotation.shape)
        ]
        if not any(label.is_anomalous for label in gt_item.get_shapes_labels()) or len(local_annotations) > 0:
            local_idx.append(idx)
    return local_idx


def get_local_subset(
    dataset: DatasetEntity,
    fully_annotated_idx: Optional[List[int]] = None,
    include_normal: bool = True,
) -> DatasetEntity:
    """Extract a subset that contains only those dataset items that have local annotations.

    Args:
        dataset (DatasetEntity): Dataset from which we want to extract the locally annotated subset.
        fully_annotated_idx (Optional[List[int]]): The indices of the fully annotated dataset items. If not provided,
            the function will compute the indices before creating the subset.
        include_normal (bool): When true, global normal annotations will be included in the local dataset.

    Returns:
        DatasetEntity: Output dataset with only local annotations
    """
    local_items = []
    if fully_annotated_idx is None:
        fully_annotated_idx = get_fully_annotated_idx(dataset)
    for idx in fully_annotated_idx:
        item = dataset[idx]

        local_annotations = [
            annotation for annotation in item.get_annotations() if not Rectangle.is_full_box(annotation.shape)
        ]
        # annotations with the normal label are considered local
        if include_normal:
            local_annotations.extend(
                [
                    annotation
                    for annotation in item.get_annotations()
                    if not any(label.label.is_anomalous for label in annotation.get_labels())
                ]
            )
        local_items.append(
            DatasetItemEntity(
                media=item.media,
                annotation_scene=AnnotationSceneEntity(
                    local_annotations,
                    kind=item.annotation_scene.kind,
                ),
                metadata=item.get_metadata(),
                subset=item.subset,
                roi=item.roi,
                ignored_labels=item.ignored_labels,
            )
        )
    return DatasetEntity(local_items, purpose=dataset.purpose)


def get_global_subset(dataset: DatasetEntity) -> DatasetEntity:
    """Extract a subset that contains only the global annotations.

    Args:
        dataset (DatasetEntity): Dataset from which we want to extract the globally annotated subset.

    Returns:
        DatasetEntity: Output dataset with only global annotations
    """
    global_items = []
    for item in dataset:
        global_annotations = [
            annotation for annotation in item.get_annotations() if Rectangle.is_full_box(annotation.shape)
        ]
        global_items.append(
            DatasetItemEntity(
                media=item.media,
                annotation_scene=AnnotationSceneEntity(global_annotations, kind=item.annotation_scene.kind),
                metadata=item.get_metadata(),
                subset=item.subset,
                roi=item.roi,
                ignored_labels=item.ignored_labels,
            )
        )
    return DatasetEntity(global_items, purpose=dataset.purpose)


def split_local_global_dataset(
    dataset: DatasetEntity,
) -> Tuple[DatasetEntity, DatasetEntity]:
    """Split a dataset into the globally and locally annotated subsets.

    Args:
        dataset (DatasetEntity): Input dataset

    Returns:
        Tuple[DatasetEntity, DatasetEntity]: Tuple of the globally and locally annotated subsets.
    """
    global_dataset = get_global_subset(dataset)
    local_dataset = get_local_subset(dataset)
    return global_dataset, local_dataset


def split_local_global_resultset(
    resultset: ResultSetEntity,
) -> Tuple[ResultSetEntity, ResultSetEntity]:
    """Split a resultset into the globally and locally annotated resultsets.

    Args:
        resultset (ResultSetEntity): Input resultset

    Returns:
        Tuple[ResultSetEntity, ResultSetEntity]: Tuple of the globally and locally annotated resultsets.
    """
    global_gt_dataset = get_global_subset(resultset.ground_truth_dataset)
    local_gt_dataset = get_local_subset(resultset.ground_truth_dataset, include_normal=False)
    local_idx = get_fully_annotated_idx(resultset.ground_truth_dataset)
    global_pred_dataset = get_global_subset(resultset.prediction_dataset)
    local_pred_dataset = get_local_subset(resultset.prediction_dataset, local_idx, include_normal=False)

    global_resultset = ResultSetEntity(
        model=resultset.model,
        ground_truth_dataset=global_gt_dataset,
        prediction_dataset=global_pred_dataset,
        purpose=resultset.purpose,
    )
    local_resultset = ResultSetEntity(
        model=resultset.model,
        ground_truth_dataset=local_gt_dataset,
        prediction_dataset=local_pred_dataset,
        purpose=resultset.purpose,
    )
    return global_resultset, local_resultset


def contains_anomalous_images(dataset: DatasetEntity) -> bool:
    """Check if a dataset contains any items with the anomalous label.

    Args:
        dataset (DatasetEntity): Dataset to check for anomalous items.

    Returns:
        bool: True if the dataset contains anomalous items, False otherwise.
    """
    for item in dataset:
        labels = item.get_shapes_labels()
        if any(label.is_anomalous for label in labels):
            return True
    return False


# pylint: disable-msg=too-many-locals
def add_saliency_maps_to_dataset_item(
    dataset_item: DatasetItemEntity,
    saliency_map: np.ndarray,
    model: Optional[ModelEntity],
    labels: List[LabelEntity],
    predicted_scored_labels: Optional[List[ScoredLabel]] = None,
    explain_predicted_classes: bool = True,
    process_saliency_maps: bool = False,
):
    """Add saliency maps(2d for class-ignore saliency map, 3d for class-wise saliency maps) to a single dataset item."""
    if saliency_map.ndim == 2:
        # Single saliency map per image, support e.g. EigenCAM use case
        if process_saliency_maps:
            saliency_map = get_actmap(saliency_map, (dataset_item.width, dataset_item.height))
        saliency_media = ResultMediaEntity(
            name="Saliency Map",
            type="saliency_map",
            annotation_scene=dataset_item.annotation_scene,
            numpy=saliency_map,
            roi=dataset_item.roi,
        )
        dataset_item.append_metadata_item(saliency_media, model=model)
    elif saliency_map.ndim == 3:
        # Multiple saliency maps per image (class-wise saliency map)
        if explain_predicted_classes:
            # Explain only predicted classes
            if predicted_scored_labels is None:
                raise ValueError("To explain only predictions, list of predicted scored labels have to be provided.")

            explain_targets = set()
            for scored_label in predicted_scored_labels:
                if scored_label.label is not None:  # Check for an empty label
                    explain_targets.add(scored_label.label)
        else:
            # Explain all classes
            explain_targets = set(labels)

        for class_id, class_wise_saliency_map in enumerate(saliency_map):
            label = labels[class_id]
            if label in explain_targets:
                if process_saliency_maps:
                    class_wise_saliency_map = get_actmap(
                        class_wise_saliency_map, (dataset_item.width, dataset_item.height)
                    )
                saliency_media = ResultMediaEntity(
                    name=label.name,
                    type="saliency_map",
                    annotation_scene=dataset_item.annotation_scene,
                    numpy=class_wise_saliency_map,
                    roi=dataset_item.roi,
                    label=label,
                )
                dataset_item.append_metadata_item(saliency_media, model=model)
    else:
        raise RuntimeError(f"Single saliency map has to be 2 or 3-dimensional, but got {saliency_map.ndim} dims")



# Copyright (C) 2018-2021 OpenMMLab
# SPDX-License-Identifier: Apache-2.0
#
# Copyright (C) 2020-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import warnings

import mmcv
import numpy as np
import torch
from mmdet.core.anchor.anchor_generator import AnchorGenerator
from mmdet.core.anchor.builder import PRIOR_GENERATORS
from torch.nn.modules.utils import _pair


@PRIOR_GENERATORS.register_module()
class SSDAnchorGeneratorClustered(AnchorGenerator):
    def __init__(self, strides, widths, heights, reclustering_anchors=False):
        self.strides = [_pair(stride) for stride in strides]
        self.widths = widths
        self.heights = heights
        self.centers = [(stride / 2.0, stride / 2.0) for stride in strides]

        self.center_offset = 0
        self.base_anchors = self.gen_base_anchors()

    def gen_base_anchors(self):
        multi_level_base_anchors = []
        for widths, heights, centers in zip(self.widths, self.heights, self.centers):
            base_anchors = self.gen_single_level_base_anchors(
                ws=torch.Tensor(widths), hs=torch.Tensor(heights), center=torch.Tensor(centers)
            )
            multi_level_base_anchors.append(base_anchors)
        return multi_level_base_anchors

    def gen_single_level_base_anchors(self, ws, hs, center):
        x_center, y_center = center

        # use float anchor and the anchor's center is aligned with the
        # pixel center
        base_anchors = [x_center - 0.5 * ws, y_center - 0.5 * hs, x_center + 0.5 * ws, y_center + 0.5 * hs]
        base_anchors = torch.stack(base_anchors, dim=-1)

        return base_anchors

    def __repr__(self):
        """str: a string that describes the module"""
        indent_str = "    "
        repr_str = self.__class__.__name__ + "(\n"
        repr_str += f"{indent_str}strides={self.strides},\n"
        repr_str += f"{indent_str}widths={self.widths},\n"
        repr_str += f"{indent_str}heights={self.heights},\n"
        repr_str += f"{indent_str}num_levels={self.num_levels}\n"
        repr_str += f"{indent_str}centers={self.centers},\n"
        repr_str += f"{indent_str}center_offset={self.center_offset})"
        return repr_str



"""This module contains the keys that can be used to retrieve parameter metadata."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from typing import List

DEFAULT_VALUE = "default_value"
MIN_VALUE = "min_value"
MAX_VALUE = "max_value"
DESCRIPTION = "description"
HEADER = "header"
WARNING = "warning"
EDITABLE = "editable"
VISIBLE_IN_UI = "visible_in_ui"
AFFECTS_OUTCOME_OF = "affects_outcome_of"
UI_RULES = "ui_rules"
TYPE = "type"
OPTIONS = "options"
ENUM_NAME = "enum_name"
AUTO_HPO_STATE = "auto_hpo_state"
AUTO_HPO_VALUE = "auto_hpo_value"


def allows_model_template_override(keyword: str) -> bool:
    """Returns True if the metadata element described by `keyword` can be overridden in a model template file.

    Args:
        keyword (str): Name of the metadata key to check.

    Returns:
        bool: True if the metadata indicated by `keyword` can be overridden in a model template .yaml file, False
        otherwise.
    """
    overrideable_keys = [
        DEFAULT_VALUE,
        MIN_VALUE,
        MAX_VALUE,
        DESCRIPTION,
        HEADER,
        EDITABLE,
        WARNING,
        VISIBLE_IN_UI,
        OPTIONS,
        ENUM_NAME,
        UI_RULES,
        AFFECTS_OUTCOME_OF,
        AUTO_HPO_STATE,
    ]
    return keyword in overrideable_keys


def allows_dictionary_values(keyword: str) -> bool:
    """Returns True if the metadata element described by `keyword` allows having a dictionary as its value.

    Args:
        keyword (str): Name of the metadata key to check.

    Returns:
        bool: True if the metadata indicated by `keyword` allows having a dictionary as its value, False otherwise.
    """
    keys_allowing_dictionary_values = [OPTIONS, UI_RULES]
    return keyword in keys_allowing_dictionary_values


def all_keys() -> List[str]:
    """Returns a list of all metadata keys.

    Returns:
        List[str]: List of all available metadata keys
    """
    return [
        DEFAULT_VALUE,
        MIN_VALUE,
        MAX_VALUE,
        DESCRIPTION,
        HEADER,
        WARNING,
        EDITABLE,
        VISIBLE_IN_UI,
        AFFECTS_OUTCOME_OF,
        UI_RULES,
        TYPE,
        OPTIONS,
        ENUM_NAME,
        AUTO_HPO_STATE,
        AUTO_HPO_VALUE,
    ]



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .inferrer import SemiSLClsInferrer
from .stage import SemiSLClsStage
from .trainer import SemiSLClsTrainer

__all__ = ["SemiSLClsStage", "SemiSLClsInferrer", "SemiSLClsTrainer"]



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcv import Config

from otx.algorithms.segmentation.adapters.mmseg.utils.config_utils import (
    patch_adaptive_repeat_dataset,
    patch_config,
    patch_datasets,
    patch_evaluation,
    patch_model_config,
    set_hyperparams,
)
from otx.algorithms.segmentation.configs.base import SegmentationConfig
from otx.api.entities.label import Domain, LabelEntity
from tests.test_suite.e2e_test_system import e2e_pytest_unit


def _create_dummy_config() -> Config:

    config: dict = dict(
        model=dict(
            type="ClassIncrEncoderDecoder",
            backbone=dict(),
            decode_head=dict(
                norm_cfg=dict(type="BN", requires_grad=True),
                loss_decode=[dict(type="CrossEntropyLoss")],
            ),
        ),
        data=dict(
            train=dict(type="RepeatDataset", times=1, adaptive_repeat=True, dataset=dict(pipeline=[])),
            val=dict(type="", pipeline=[]),
            test=dict(type="", pipeline=[]),
        ),
        runner=dict(type="EpochRunnerWithCancel", max_epochs=300),
        checkpoint_config=dict(interval=1),
        optimizer=dict(),
        lr_config=dict(policy="ReduceLROnPlateau", warmup_iters=80, fixed_iters=100),
        evaluation=dict(interval=1, metric=["mIoU", "mDice"]),
        custom_hooks=[],
        params_config=dict(iters=0),
    )

    return Config(config)


class TestConfigUtilsValidation:
    @e2e_pytest_unit
    def test_patch_config(self) -> None:
        config: Config = _create_dummy_config()
        labels: list[LabelEntity] = [LabelEntity(name="test label", domain=Domain.SEGMENTATION)]
        num_classes: int = len(labels) + 1
        work_dir: str = "./work_dir"
        patch_config(config, work_dir, labels)

        assert "custom_hooks" in config
        assert "train_pipeline" not in config
        assert config.work_dir == work_dir
        assert len(config.data["train"].classes) == num_classes

    @e2e_pytest_unit
    def test_patch_model_config_nondistributed(self) -> None:
        config: Config = _create_dummy_config()
        labels: list[LabelEntity] = [LabelEntity(name="test label", domain=Domain.SEGMENTATION)]
        num_classes: int = len(labels) + 1
        distributed: bool = False
        patch_model_config(config, labels, distributed)

        assert config.model.decode_head["num_classes"] == num_classes
        assert config.model.decode_head.norm_cfg.type != "SyncBN"

    @e2e_pytest_unit
    def test_patch_datasets(self) -> None:
        config: Config = _create_dummy_config()
        patch_datasets(config, type="MPASegDataset")

        assert "classes" not in config.data.train.dataset
        assert "labels" in config.data.train.dataset

    @e2e_pytest_unit
    def test_patch_evaluation(self) -> None:
        config: Config = _create_dummy_config()
        patch_evaluation(config)

        assert config.early_stop_metric == "mDice"

    @e2e_pytest_unit
    def test_set_hyperparams(self) -> None:
        config: Config = _create_dummy_config()
        hyperparameters: SegmentationConfig = SegmentationConfig()
        set_hyperparams(config, hyperparameters)

        assert config.runner.max_epochs > 0

    @e2e_pytest_unit
    def test_patch_adaptive_repeat_dataset(self) -> None:
        config: Config = _create_dummy_config()
        num_samples: int = 10

        initial_epochs: int = config.runner.max_epochs
        patch_adaptive_repeat_dataset(config, num_samples)
        epochs_after_adaptive_repeat: int = config.runner.max_epochs

        assert initial_epochs > epochs_after_adaptive_repeat



"""Base configuration parameters for Draem."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


from .configuration import DraemAnomalyBaseConfig

__all__ = ["DraemAnomalyBaseConfig"]



"""This module implements utilities for labels."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Optional

from otx.api.entities.label import LabelEntity
from otx.api.entities.label_schema import LabelSchemaEntity


def get_empty_label(label_schema: LabelSchemaEntity) -> Optional[LabelEntity]:
    """Get first empty label from label_schema."""
    empty_candidates = list(
        set(label_schema.get_labels(include_empty=True)) - set(label_schema.get_labels(include_empty=False))
    )
    if empty_candidates:
        return empty_candidates[0]
    return None



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from collections.abc import Sequence
from copy import deepcopy

import numpy as np
from mmcv.utils import build_from_cfg
from mmseg.datasets.builder import PIPELINES
from scipy.ndimage import gaussian_filter


@PIPELINES.register_module()
class ProbCompose(object):
    def __init__(self, transforms, probs):
        assert isinstance(transforms, Sequence)
        assert isinstance(probs, Sequence)
        assert len(transforms) == len(probs)
        assert all(p >= 0.0 for p in probs)

        sum_probs = float(sum(probs))
        assert sum_probs > 0.0
        norm_probs = [float(p) / sum_probs for p in probs]
        self.limits = np.cumsum([0.0] + norm_probs)

        self.transforms = []
        for transform in transforms:
            if isinstance(transform, dict):
                transform = build_from_cfg(transform, PIPELINES)
                self.transforms.append(transform)
            elif callable(transform):
                self.transforms.append(transform)
            else:
                raise TypeError(f"transform must be callable or a dict, but got {type(transform)}")

    def __call__(self, data):
        rand_value = np.random.rand()
        transform_id = np.max(np.where(rand_value > self.limits)[0])

        transform = self.transforms[transform_id]
        data = transform(data)

        return data

    def __repr__(self):
        format_string = self.__class__.__name__ + "("
        for t in self.transforms:
            format_string += "\n"
            format_string += "    {0}".format(t)
        format_string += "\n)"

        return format_string


@PIPELINES.register_module()
class MaskCompose(object):
    def __init__(self, transforms, prob, lambda_limits=(4, 16), keep_original=False):
        self.keep_original = keep_original
        self.prob = prob
        assert 0.0 <= self.prob <= 1.0

        assert isinstance(lambda_limits, Sequence)
        assert len(lambda_limits) == 2
        assert 0.0 < lambda_limits[0] < lambda_limits[1]
        self.lambda_limits = lambda_limits

        assert isinstance(transforms, Sequence)
        self.transforms = []
        for transform in transforms:
            if isinstance(transform, dict):
                transform = build_from_cfg(transform, PIPELINES)
                self.transforms.append(transform)
            elif callable(transform):
                self.transforms.append(transform)
            else:
                raise TypeError(f"transform must be callable or a dict, but got {type(transform)}")

    @staticmethod
    def _apply_transforms(data, transforms):
        for t in transforms:
            data = t(data)
            if data is None:
                return None

        return data

    @staticmethod
    def _generate_mask(shape, lambda_limits):
        noise = np.random.randn(*shape)

        sigma = np.exp(np.log10(np.random.uniform(lambda_limits[0], lambda_limits[1])))
        soft_mask = gaussian_filter(noise, sigma=sigma)

        threshold = np.median(soft_mask)
        hard_mask = soft_mask > threshold

        return hard_mask

    @staticmethod
    def _mix_img(main_img, aux_img, mask):
        return np.where(np.expand_dims(mask, axis=2), main_img, aux_img)

    def __call__(self, data):
        main_data = self._apply_transforms(deepcopy(data), self.transforms)
        assert main_data is not None
        if not self.keep_original and np.random.rand() > self.prob:
            return main_data

        aux_data = self._apply_transforms(deepcopy(data), self.transforms)
        assert aux_data is not None

        assert main_data["img"].shape == aux_data["img"].shape

        mask = self._generate_mask(main_data["img"].shape[:2], self.lambda_limits)
        mixed_img = self._mix_img(main_data["img"], aux_data["img"], mask)

        if self.keep_original:
            main_data["aux_img"] = mixed_img
        else:
            main_data["img"] = mixed_img

        return main_data

    def __repr__(self):
        format_string = self.__class__.__name__ + "("
        for t in self.transforms:
            format_string += "\n"
            format_string += "    {0}".format(t)
        format_string += "\n)"

        return format_string



"""This module contains all elements needed to construct a OTX configuration object."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from .configurable_enum import ConfigurableEnum
from .parameter_group import ParameterGroup, add_parameter_group
from .primitive_parameters import (
    boolean_attribute,
    configurable_boolean,
    configurable_float,
    configurable_integer,
    float_selectable,
    selectable,
    string_attribute,
)

__all__ = [
    "ConfigurableEnum",
    "ParameterGroup",
    "add_parameter_group",
    "boolean_attribute",
    "configurable_boolean",
    "configurable_float",
    "configurable_integer",
    "float_selectable",
    "selectable",
    "string_attribute",
]



"""Test schedulers."""


# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import pytest

from otx.mpa.modules.models.scalar_schedulers import (
    ConstantScalarScheduler,
    PolyScalarScheduler,
    StepScalarScheduler,
)


class TestSchedulers:
    """Test schedulers."""

    def test_constant_scalar_scheduler(self):
        """Test constant scalar scheduler.

        Learning rate should not change over time.
        """
        scheduler = ConstantScalarScheduler(scale=30.0)
        assert scheduler(0, 1) == 30.0
        assert scheduler(1, 1) == 30.0
        assert scheduler(2, 10) == 30.0

    def test_constant_scalar_scheduler_invalid_scale(self):
        """Test constant scalar scheduler with invalid scale."""
        with pytest.raises(AssertionError):
            ConstantScalarScheduler(scale=-1.0)

    @pytest.mark.xfail
    def test_constant_scalar_scheduler_invalid_step(self):
        """Test constant scalar scheduler with invalid step.

        TODO: ConstantScalarScheculer should be modified to raise this error
        """
        scheduler = ConstantScalarScheduler(scale=30.0)
        with pytest.raises(AssertionError):
            scheduler(-1, 1)

    def test_poly_scalar_scheduler_by_epoch_false(self):
        """Test poly scalar scheduler."""
        # By epoch is False
        scheduler = PolyScalarScheduler(
            start_scale=30.0,
            end_scale=0.0,
            num_iters=100,
            power=0.9,
            by_epoch=False,
        )

        # learning rate should decrease over time
        assert scheduler(0, 1) == 30.0
        assert scheduler(1, 1) < 30.0
        assert scheduler(2, 1) < scheduler(1, 1)
        assert scheduler(3, 1) < scheduler(2, 1)

        assert scheduler(50, 10) == scheduler(50, 1)  # as this is not by epoch

        # learning rate should not change after num_iters
        assert scheduler(100, 1) == 0.0
        assert scheduler(101, 1) == 0.0
        assert scheduler(102, 1) == 0.0

    def test_poly_scalar_scheduler_by_epoch_true(self):
        scheduler = PolyScalarScheduler(
            start_scale=30.0,
            end_scale=0.0,
            num_iters=100,
            power=0.9,
            by_epoch=True,
        )

        # learning rate should decrease over time
        assert scheduler(0, 1) == 30.0
        assert scheduler(1, 1) < 30.0
        assert scheduler(2, 1) < scheduler(1, 1)
        assert scheduler(3, 1) < scheduler(2, 1)

        assert scheduler(50, 10) != scheduler(50, 1)  # as this is by epoch

        # learning rate should not change after num_iters
        assert scheduler(100, 1) == 0.0
        assert scheduler(101, 1) == 0.0
        assert scheduler(102, 1) == 0.0

    def test_step_scalar_scheduler_by_epoch_false(self):
        """Test step scalar scheduler."""
        # By epoch is False
        scheduler = StepScalarScheduler(
            scales=[30.0, 20.0, 10.0, 5.0],
            num_iters=[2, 3, 4],
            by_epoch=False,
        )

        # learning rate should decrease over time as a step function
        assert scheduler(0, 1) == 30.0
        assert scheduler(1, 1) == 30.0
        assert scheduler(2, 1) < scheduler(1, 1)
        assert scheduler(3, 1) < scheduler(2, 1)

        assert scheduler(50, 10) == scheduler(50, 1)

        assert scheduler(5, 2) == 5.0
        assert scheduler(5, 0) == scheduler(10, 1)

        assert scheduler(10, 1) == 5.0  # steps greater than total num_iters

    def test_step_scalar_scheduler_by_epoch_true(self):
        # By epoch is True
        scheduler = StepScalarScheduler(
            scales=[30.0, 20.0, 10.0, 5.0],
            num_iters=[2, 3, 4],
            by_epoch=True,
        )

        # learning rate should decrease over time as a step function
        assert scheduler(0, 1) == 30.0
        assert scheduler(1, 1) == 30.0
        assert scheduler(2, 1) < scheduler(1, 1)
        assert scheduler(3, 1) < scheduler(2, 1)

        assert scheduler(9, 5) == 30.0
        assert scheduler(5, 2) == 20.0
        assert scheduler(5, 2) < scheduler(10, 11)



"""HPO runner and resource manager class."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import logging
import multiprocessing
import os
from copy import deepcopy
from functools import partial
from typing import Any, Callable, Dict, Literal, Optional, Union

from otx.hpo.hpo_base import HpoBase, Trial, TrialStatus
from otx.hpo.resource_manager import get_resource_manager

logger = logging.getLogger(__name__)


class HpoLoop:
    """HPO loop manager to run trials.

    Args:
        hpo_algo (HpoBase): HPO algorithms.
        train_func (Callable): Function to train a model.
        resource_type (Literal['gpu', 'cpu'], optional): Which type of resource to use.
                                                         If can be changed depending on environment. Defaults to "gpu".
        num_parallel_trial (Optional[int], optional): How many trials to run in parallel.
                                                    It's used for CPUResourceManager. Defaults to None.
        num_gpu_for_single_trial (Optional[int], optional): How many GPUs are used for a single trial.
                                                            It's used for GPUResourceManager. Defaults to None.
        available_gpu (Optional[str], optional): How many GPUs are available. It's used for GPUResourceManager.
                                                 Defaults to None.
    """

    def __init__(
        self,
        hpo_algo: HpoBase,
        train_func: Callable,
        resource_type: Literal["gpu", "cpu"] = "gpu",
        num_parallel_trial: Optional[int] = None,
        num_gpu_for_single_trial: Optional[int] = None,
        available_gpu: Optional[str] = None,
    ):
        self._hpo_algo = hpo_algo
        self._train_func = train_func
        self._running_trials: Dict[int, Dict] = {}
        self._mp = multiprocessing.get_context("spawn")
        self._uid_index = 0
        self._resource_manager = get_resource_manager(
            resource_type, num_parallel_trial, num_gpu_for_single_trial, available_gpu
        )

    def run(self):
        """Run a HPO loop."""
        logger.info("HPO loop starts.")
        while not self._hpo_algo.is_done():
            if self._resource_manager.have_available_resource():
                trial = self._hpo_algo.get_next_sample()
                if trial is not None:
                    self._start_trial_process(trial)

            self._remove_finished_process()
            self._get_reports()

        logger.info("HPO loop is done.")
        self._get_reports()
        self._join_all_processes()

    def _start_trial_process(self, trial: Trial):
        logger.info(f"{trial.id} trial is now running.")
        logger.debug(f"{trial.id} hyper paramter => {trial.configuration}")

        trial.status = TrialStatus.RUNNING
        uid = self._get_uid()

        origin_env = deepcopy(os.environ)
        env = self._resource_manager.reserve_resource(uid)
        if env is not None:
            for key, val in env.items():
                os.environ[key] = val

        pipe1, pipe2 = self._mp.Pipe(True)
        process = self._mp.Process(
            target=_run_train,
            args=(
                self._train_func,
                trial.get_train_configuration(),
                partial(_report_score, pipe=pipe2, trial_id=trial.id),
            ),
        )
        os.environ = origin_env
        self._running_trials[uid] = {"process": process, "trial": trial, "pipe": pipe1}
        process.start()

    def _remove_finished_process(self):
        trial_to_remove = []
        for uid, val in self._running_trials.items():
            process = val["process"]
            if not process.is_alive():
                val["pipe"].close()
                process.join()
                trial_to_remove.append(uid)

        for uid in trial_to_remove:
            trial = self._running_trials[uid]["trial"]
            trial.status = TrialStatus.STOP
            self._resource_manager.release_resource(uid)
            del self._running_trials[uid]

    def _get_reports(self):
        for trial in self._running_trials.values():
            pipe = trial["pipe"]
            if pipe.poll():
                try:
                    report = pipe.recv()
                except EOFError:
                    continue
                trial_status = self._hpo_algo.report_score(
                    report["score"], report["progress"], report["trial_id"], report["done"]
                )
                pipe.send(trial_status)

        self._hpo_algo.save_results()

    def _join_all_processes(self):
        for val in self._running_trials.values():
            val["pipe"].close()

        for val in self._running_trials.values():
            process = val["process"]
            process.join()

        self._running_trials = {}

    def _get_uid(self) -> int:
        uid = self._uid_index
        self._uid_index += 1
        return uid


def _run_train(train_func: Callable, hp_config: Dict, report_func: Callable):
    # set multi process method as default
    multiprocessing.set_start_method(None, True)  # type: ignore
    train_func(hp_config, report_func)


def _report_score(score: Union[int, float], progress: Union[int, float], pipe, trial_id: Any, done: bool = False):
    logger.debug(f"score : {score}, progress : {progress}, trial_id : {trial_id}, pid : {os.getpid()}, done : {done}")
    try:
        pipe.send({"score": score, "progress": progress, "trial_id": trial_id, "pid": os.getpid(), "done": done})
    except BrokenPipeError:
        return TrialStatus.STOP
    try:
        trial_status = pipe.recv()
    except EOFError:
        return TrialStatus.STOP

    logger.debug(f"trial_status : {trial_status}")
    return trial_status


def run_hpo_loop(
    hpo_algo: HpoBase,
    train_func: Callable,
    resource_type: Literal["gpu", "cpu"] = "gpu",
    num_parallel_trial: Optional[int] = None,
    num_gpu_for_single_trial: Optional[int] = None,
    available_gpu: Optional[str] = None,
):
    """Run the HPO loop.

    Args:
        hpo_algo (HpoBase): HPO algorithms.
        train_func (Callable): Function to train a model.
        resource_type (Literal['gpu', 'cpu'], optional): Which type of resource to use.
                                                         If can be changed depending on environment. Defaults to "gpu".
        num_parallel_trial (Optional[int], optional): How many trials to run in parallel.
                                                      It's used for CPUResourceManager. Defaults to None.
        num_gpu_for_single_trial (Optional[int], optional): How many GPUs are used for a single trial.
                                                            It's used for GPUResourceManager. Defaults to None.
        available_gpu (Optional[str], optional): How many GPUs are available. It's used for GPUResourceManager.
                                                 Defaults to None.
    """
    hpo_loop = HpoLoop(hpo_algo, train_func, resource_type, num_parallel_trial, num_gpu_for_single_trial, available_gpu)
    hpo_loop.run()



"""MMDeploy config of OCR-Lite-HRnet-18 model for Segmentation Task."""

_base_ = ["../base/deployments/base_segmentation_dynamic.py"]

ir_config = dict(
    output_names=["output", "feature_vector"],
)

backend_config = dict(
    model_inputs=[dict(opt_shapes=dict(input=[-1, 3, 544, 544]))],
)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from abc import abstractmethod

import torch
import torch.nn.functional as F
from mmseg.models.losses.utils import weight_reduce_loss

from otx.mpa.modules.models.builder import build_scalar_scheduler

from .base_weighted_loss import BaseWeightedLoss


def entropy(p, dim=1, keepdim=False):
    return -torch.where(p > 0.0, p * p.log(), torch.zeros_like(p)).sum(dim=dim, keepdim=keepdim)


class BasePixelLoss(BaseWeightedLoss):
    def __init__(self, scale_cfg=None, pr_product=False, conf_penalty_weight=None, border_reweighting=False, **kwargs):
        super(BasePixelLoss, self).__init__(**kwargs)

        self._enable_pr_product = pr_product
        self._border_reweighting = border_reweighting

        self._reg_weight_scheduler = build_scalar_scheduler(conf_penalty_weight)
        self._scale_scheduler = build_scalar_scheduler(scale_cfg, default_value=1.0)

        self._last_scale = 0.0
        self._last_reg_weight = 0.0

    @property
    def last_scale(self):
        return self._last_scale

    @property
    def last_reg_weight(self):
        return self._last_reg_weight

    @property
    def with_regularization(self):
        return self._reg_weight_scheduler is not None

    @property
    def with_pr_product(self):
        return self._enable_pr_product

    @property
    def with_border_reweighting(self):
        return self._border_reweighting

    @staticmethod
    def _pr_product(prod):
        alpha = torch.sqrt(1.0 - prod.pow(2.0))
        out_prod = alpha.detach() * prod + prod.detach() * (1.0 - alpha)

        return out_prod

    @staticmethod
    def _regularization(logits, scale, weight):
        probs = F.softmax(scale * logits, dim=1)
        entropy_values = entropy(probs, dim=1)
        out_values = -weight * entropy_values

        return out_values

    @staticmethod
    def _sparsity(values, valid_mask):
        with torch.no_grad():
            valid_values = values[valid_mask]
            sparsity = 1.0 - valid_values.count_nonzero() / max(1.0, valid_mask.sum())
        return sparsity.item()

    @staticmethod
    def _pred_stat(output, labels, valid_mask, window_size=5, min_group_ratio=0.6):
        assert window_size > 1
        assert 0.0 < min_group_ratio < 1.0

        min_group_size = int(min_group_ratio * window_size * window_size)
        assert min_group_size > 0

        with torch.no_grad():
            predictions = torch.argmax(output, dim=1)
            invalid_pred_mask = valid_mask & (predictions != labels)

            group_sizes = F.avg_pool2d(
                invalid_pred_mask.float(),
                kernel_size=window_size,
                stride=1,
                padding=(window_size - 1) // 2,
                divisor_override=1,
            )
            large_group_mask = invalid_pred_mask & (group_sizes >= min_group_size)

            num_target = torch.sum(large_group_mask, dim=(1, 2))
            num_total = torch.sum(invalid_pred_mask, dim=(1, 2))
            out_ratio = torch.mean(num_target / num_total.clamp_min(1))

        return out_ratio.item()

    def _forward(self, output, labels, avg_factor=None, pixel_weights=None, reduction_override=None):
        assert reduction_override in (None, "none", "mean", "sum")
        reduction = reduction_override if reduction_override else self.reduction

        self._last_scale = self._scale_scheduler(self.iter, self.epoch_size)

        if self.with_pr_product:
            output = self._pr_product(output)

        num_classes = output.size(1)
        valid_labels = torch.clamp(labels, 0, num_classes - 1)
        valid_mask = labels != self.ignore_index

        losses, updated_output = self._calculate(output, valid_labels, self._last_scale)

        if self.with_regularization:
            self._last_reg_weight = self._reg_weight_scheduler(self.iter, self.epoch_size)
            regularization = self._regularization(updated_output, self._last_scale, self._last_reg_weight)
            losses = torch.clamp_min(losses + regularization, 0.0)

        if self.with_border_reweighting:
            assert pixel_weights is not None
            losses = pixel_weights.squeeze(1) * losses

        losses = torch.where(valid_mask, losses, torch.zeros_like(losses))
        raw_sparsity = self._sparsity(losses, valid_mask)
        invalid_ratio = self._pred_stat(output, labels, valid_mask)

        weight, weight_sparsity = None, 0.0
        if self.sampler is not None:
            weight = self.sampler(losses, output, valid_labels, valid_mask)
            weight_sparsity = self._sparsity(weight, valid_mask)

        loss = weight_reduce_loss(losses, weight=weight, reduction=reduction, avg_factor=avg_factor)

        meta = dict(
            weight=self.last_loss_weight,
            reg_weight=self.last_reg_weight,
            scale=self.last_scale,
            raw_sparsity=raw_sparsity,
            weight_sparsity=weight_sparsity,
            invalid_ratio=invalid_ratio,
        )

        return loss, meta

    @abstractmethod
    def _calculate(self, output, labels, scale):
        pass



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


class LossEqualizer:
    def __init__(self, weights=None, momentum=0.1):
        self.momentum = momentum

        self.trg_ratios = None
        if weights is not None:
            assert isinstance(weights, dict)
            assert len(weights) > 0

            sum_weight = 0.0
            for loss_weight in weights.values():
                assert loss_weight > 0
                sum_weight += float(loss_weight)
            assert sum_weight > 0.0

            self.trg_ratios = {loss_name: float(loss_weight) / sum_weight for loss_name, loss_weight in weights.items()}

        self._smoothed_values = dict()

    def reweight(self, losses):
        assert isinstance(losses, dict)

        if len(losses) == 0:
            return losses

        for loss_name, loss_value in losses.items():
            if loss_name not in self._smoothed_values:
                self._smoothed_values[loss_name] = loss_value.item()
            else:
                smoothed_loss = self._smoothed_values[loss_name]
                self._smoothed_values[loss_name] = (
                    1.0 - self.momentum
                ) * smoothed_loss + self.momentum * loss_value.item()

        if len(self._smoothed_values) == 1:
            return losses

        total_sum = sum(self._smoothed_values.values())
        trg_value_default = total_sum / float(len(self._smoothed_values))

        weighted_losses = dict()
        for loss_name, loss_value in losses.items():
            if self.trg_ratios is not None:
                assert loss_name in self.trg_ratios.keys()
                trg_value = self.trg_ratios[loss_name] * total_sum
            else:
                trg_value = trg_value_default

            loss_weight = trg_value / self._smoothed_values[loss_name]
            weighted_losses[loss_name] = loss_weight * loss_value

        return weighted_losses



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from dataclasses import dataclass, field

import torch

from .builder import OPS
from .op import Attribute, Operation


@dataclass
class ReduceMeanV1Attribute(Attribute):
    keep_dims: bool = field(default=False)


@OPS.register()
class ReduceMeanV1(Operation[ReduceMeanV1Attribute]):
    TYPE = "ReduceMean"
    VERSION = 1
    ATTRIBUTE_FACTORY = ReduceMeanV1Attribute

    def forward(self, input, axes):
        if isinstance(axes, torch.Tensor):
            axes = axes.tolist()
        if not axes:
            return input

        if not isinstance(axes, (list, tuple)):
            axes = [axes]

        return torch.mean(input=input, dim=axes, keepdim=self.attrs.keep_dims)


@dataclass
class ReduceProdV1Attribute(Attribute):
    keep_dims: bool = field(default=False)


@OPS.register()
class ReduceProdV1(Operation[ReduceProdV1Attribute]):
    TYPE = "ReduceProd"
    VERSION = 1
    ATTRIBUTE_FACTORY = ReduceProdV1Attribute

    def forward(self, input, axes):
        if isinstance(axes, torch.Tensor):
            axes = axes.tolist()
        if not axes:
            return input

        if not isinstance(axes, (list, tuple)):
            axes = [axes]

        output = input
        for ax in axes:
            output = torch.prod(input=output, dim=ax, keepdim=True)
        if not self.attrs.keep_dims:
            output = torch.squeeze(output)

        return output


@dataclass
class ReduceMinV1Attribute(Attribute):
    keep_dims: bool = field(default=False)


@OPS.register()
class ReduceMinV1(Operation[ReduceMinV1Attribute]):
    TYPE = "ReduceMin"
    VERSION = 1
    ATTRIBUTE_FACTORY = ReduceMinV1Attribute

    def forward(self, input, axes):
        if isinstance(axes, torch.Tensor):
            axes = axes.tolist()
        if not axes:
            return input

        if not isinstance(axes, (list, tuple)):
            axes = [axes]

        output = input
        for ax in axes:
            output = torch.min(input=output, dim=ax, keepdim=True)[0]
        if not self.attrs.keep_dims:
            output = torch.squeeze(output)

        return output


@dataclass
class ReduceSumV1Attribute(Attribute):
    keep_dims: bool = field(default=False)


@OPS.register()
class ReduceSumV1(Operation[ReduceSumV1Attribute]):
    TYPE = "ReduceSum"
    VERSION = 1
    ATTRIBUTE_FACTORY = ReduceSumV1Attribute

    def forward(self, input, axes):
        if isinstance(axes, torch.Tensor):
            axes = axes.tolist()
        if not axes:
            return input

        return torch.sum(input=input, dim=axes, keepdim=self.attrs.keep_dims)



"""Segmentation Dataset Adapter."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import json

# pylint: disable=invalid-name, too-many-locals, no-member, too-many-nested-blocks, too-many-branches
import os
from typing import Dict, List, Optional

import cv2
import numpy as np
from datumaro.components.annotation import AnnotationType, Mask
from datumaro.components.dataset import Dataset as DatumaroDataset
from datumaro.plugins.data_formats.common_semantic_segmentation import (
    CommonSemanticSegmentationBase,
    make_categories,
)
from datumaro.plugins.transforms import MasksToPolygons
from datumaro.util.meta_file_util import parse_meta_file
from skimage.segmentation import felzenszwalb

from otx.api.entities.annotation import Annotation
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.image import Image
from otx.api.entities.model_template import TaskType
from otx.api.entities.subset import Subset
from otx.core.data.adapter.base_dataset_adapter import BaseDatasetAdapter
from otx.mpa.utils.logger import get_logger


class SegmentationDatasetAdapter(BaseDatasetAdapter):
    """Segmentation adapter inherited from BaseDatasetAdapter.

    It converts DatumaroDataset --> DatasetEntity for semantic segmentation task
    """

    def __init__(
        self,
        task_type: TaskType,
        train_data_roots: Optional[str] = None,
        val_data_roots: Optional[str] = None,
        test_data_roots: Optional[str] = None,
        unlabeled_data_roots: Optional[str] = None,
    ):
        super().__init__(task_type, train_data_roots, val_data_roots, test_data_roots, unlabeled_data_roots)
        self.updated_label_id: Dict[int, int] = {}

    def get_otx_dataset(self) -> DatasetEntity:
        """Convert DatumaroDataset to DatasetEntity for Segmentation."""
        # Prepare label information
        label_information = self._prepare_label_information(self.dataset)
        self.label_entities = label_information["label_entities"]

        dataset_items: List[DatasetItemEntity] = []
        used_labels: List[int] = []

        if hasattr(self, "data_type_candidates"):
            if self.data_type_candidates[0] == "voc":
                self.set_voc_labels()

            if self.data_type_candidates[0] == "common_semantic_segmentation":
                self.set_common_labels()

        else:
            # For datasets used for self-sl.
            # They are not included in any data type and `data_type_candidates` is not set,
            # so they must be handled independently. But, setting `self.updated_label_id` is compatible
            # with "common_semantic_segmentation", so we can use it.
            self.set_common_labels()

        for subset, subset_data in self.dataset.items():
            for _, datumaro_items in subset_data.subsets().items():
                for datumaro_item in datumaro_items:
                    image = Image(file_path=datumaro_item.media.path)
                    shapes: List[Annotation] = []
                    for ann in datumaro_item.annotations:
                        if ann.type == AnnotationType.mask:
                            # TODO: consider case -> didn't include the background information
                            datumaro_polygons = MasksToPolygons.convert_mask(ann)
                            for d_polygon in datumaro_polygons:
                                new_label = self.updated_label_id.get(d_polygon.label, None)
                                if new_label is not None:
                                    d_polygon.label = new_label
                                else:
                                    continue

                                shapes.append(self._get_polygon_entity(d_polygon, image.width, image.height))
                                if d_polygon.label not in used_labels:
                                    used_labels.append(d_polygon.label)

                    dataset_item = DatasetItemEntity(image, self._get_ann_scene_entity(shapes), subset=subset)
                    dataset_items.append(dataset_item)

        self.remove_unused_label_entities(used_labels)
        return DatasetEntity(items=dataset_items)

    def set_voc_labels(self):
        """Set labels for common_semantic_segmentation dataset."""
        # Remove background & ignored label in VOC from datumaro
        self._remove_labels(["background", "ignored"])

    def set_common_labels(self):
        """Set labels for common_semantic_segmentation dataset."""
        # Remove background if in label_entities
        is_removed = self._remove_labels(["background"])

        # Shift label id since datumaro always extracts bg polygon with label 0
        if is_removed is False:
            self.updated_label_id = {k + 1: v for k, v in self.updated_label_id.items()}

    def _remove_labels(self, label_names: List):
        """Remove background label in label entity set."""
        is_removed = False
        new_label_entities = []
        for i, entity in enumerate(self.label_entities):
            if entity.name not in label_names:
                new_label_entities.append(entity)
            else:
                is_removed = True

        self.label_entities = new_label_entities

        for i, entity in enumerate(self.label_entities):
            self.updated_label_id[int(entity.id)] = i
            entity.id = ID(i)

        return is_removed


class SelfSLSegmentationDatasetAdapter(SegmentationDatasetAdapter):
    """Self-SL for segmentation adapter inherited from SegmentationDatasetAdapter."""

    # pylint: disable=protected-access
    def _import_dataset(
        self,
        train_data_roots: Optional[str] = None,
        val_data_roots: Optional[str] = None,
        test_data_roots: Optional[str] = None,
        unlabeled_data_roots: Optional[str] = None,
        pseudo_mask_dir: str = "detcon_mask",
    ) -> Dict[Subset, DatumaroDataset]:
        """Import custom Self-SL dataset for using DetCon.

        Self-SL for semantic segmentation using DetCon uses pseudo masks as labels,
        but Datumaro cannot load this custom data structure because it is not in Datumaro format.
        So, it is required to manually load and set annotations.

        Args:
            train_data_roots (Optional[str]): Path for training data.
            pseudo_mask_dir (str): Directory to save pseudo masks. Defaults to "detcon_mask".

        Returns:
            DatumaroDataset: Datumaro Dataset
        """
        if train_data_roots is None:
            raise ValueError("train_data_root must be set.")

        logger = get_logger()
        logger.warning(f"Please check if {train_data_roots} is data roots only for images, not annotations.")

        dataset = {}
        dataset[Subset.TRAINING] = DatumaroDataset.import_from(train_data_roots, format="image_dir")
        self.is_train_phase = True

        # Load pseudo masks
        img_dir = None
        total_labels = []
        for item in dataset[Subset.TRAINING]:
            img_path = item.media.path
            if img_dir is None:
                # Get image directory
                img_dir = train_data_roots.split("/")[-1]
            pseudo_mask_path = img_path.replace(img_dir, pseudo_mask_dir)
            if pseudo_mask_path.endswith(".jpg"):
                pseudo_mask_path = pseudo_mask_path.replace(".jpg", ".png")

            if not os.path.isfile(pseudo_mask_path):
                # Create pseudo mask
                pseudo_mask = self.create_pseudo_masks(item.media.data, pseudo_mask_path)  # type: ignore
            else:
                # Load created pseudo mask
                pseudo_mask = cv2.imread(pseudo_mask_path, cv2.IMREAD_GRAYSCALE)

            # Set annotations into each item
            annotations = []
            labels = np.unique(pseudo_mask)
            for label_id in labels:
                if label_id not in total_labels:
                    # Stack label_id to save dataset_meta.json
                    total_labels.append(label_id)
                annotations.append(
                    Mask(image=CommonSemanticSegmentationBase._lazy_extract_mask(pseudo_mask, label_id), label=label_id)
                )
            item.annotations = annotations

        pseudo_mask_roots = train_data_roots.replace(img_dir, pseudo_mask_dir)  # type: ignore
        if not os.path.isfile(os.path.join(pseudo_mask_roots, "dataset_meta.json")):
            # Save dataset_meta.json for newly created pseudo masks
            # FIXME: Because background class is ignored when generating polygons, meta is set with len(labels)-1.
            # It must be considered to set the whole labels later.
            # (-> {i: f"target{i+1}" for i in range(max(total_labels)+1)})
            meta = {"label_map": {i + 1: f"target{i+1}" for i in range(max(total_labels))}}
            with open(os.path.join(pseudo_mask_roots, "dataset_meta.json"), "w", encoding="UTF-8") as f:
                json.dump(meta, f, indent=4)

        # Make categories for pseudo masks
        label_map = parse_meta_file(os.path.join(pseudo_mask_roots, "dataset_meta.json"))
        dataset[Subset.TRAINING].define_categories(make_categories(label_map))

        return dataset

    def create_pseudo_masks(self, img: np.array, pseudo_mask_path: str, mode: str = "FH") -> None:
        """Create pseudo masks for self-sl for semantic segmentation using DetCon.

        Args:
            img (np.array) : A sample to create a pseudo mask.
            pseudo_mask_path (str): The path to save a pseudo mask.
            mode (str): The mode to create a pseudo mask. Defaults to "FH".

        Returns:
            np.array: a created pseudo mask for item.
        """
        if mode == "FH":
            pseudo_mask = felzenszwalb(img, scale=1000, min_size=1000)
        else:
            raise ValueError((f'{mode} is not supported to create pseudo masks for DetCon. Choose one of ["FH"].'))

        os.makedirs(os.path.dirname(pseudo_mask_path), exist_ok=True)
        cv2.imwrite(pseudo_mask_path, pseudo_mask.astype(np.uint8))

        return pseudo_mask



"""Wrapper for Open Model Zoo for Anomaly Detection tasks."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Any, Dict

import cv2
import numpy as np

from .base import AnomalyBase


class AnomalyDetection(AnomalyBase):
    """Wrapper for anomaly detection task."""

    __model__ = "anomaly_detection"

    def postprocess(self, outputs: Dict[str, np.ndarray], meta: Dict[str, Any]) -> np.ndarray:
        """Resize the outputs of the model to original image size.

        Args:
            outputs (Dict[str, np.ndarray]): Raw outputs of the model after ``infer_sync`` is called.
            meta (Dict[str, Any]): Metadata which contains values such as threshold, original image size.

        Returns:
            np.ndarray: Detection Mask
        """
        anomaly_map: np.ndarray = outputs[self.output_blob_name].squeeze()

        meta["pixel_threshold"] = self.pixel_threshold  # pylint: disable=no-member
        meta["min"] = self.min  # pylint: disable=no-member
        meta["max"] = self.max  # pylint: disable=no-member
        meta["threshold"] = self.threshold  # pylint: disable=no-member

        anomaly_map = self._normalize(anomaly_map, meta["pixel_threshold"], meta["min"], meta["max"])

        input_image_height = meta["original_shape"][0]
        input_image_width = meta["original_shape"][1]
        result = cv2.resize(anomaly_map, (input_image_width, input_image_height))

        return result



"""MobileNet-V3-large-075 for hierarchical config."""

# pylint: disable=invalid-name

_base_ = "../base/models/mobilenet_v3.py"

model = dict(
    type="SAMImageClassifier",
    task="classification",
    backbone=dict(
        mode="large",
        width_mult=0.75,
    ),
    head=dict(
        type="CustomHierarchicalNonLinearClsHead",
        in_channels=720,
        hid_channels=1280,
        multilabel_loss=dict(
            type="AsymmetricLossWithIgnore",
            gamma_pos=0.0,
            gamma_neg=4.0,
        ),
    ),
)



import pytest
import torch
from mmcls.models.builder import build_head

from otx.mpa.modules.models.heads.semisl_cls_head import (
    SemiLinearClsHead,
    SemiNonLinearClsHead,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestSemiSLClsHead:
    @pytest.fixture(autouse=True)
    def setUp(self):
        """Semi-SL for Classification Head Settings."""
        self.in_channels = 1280
        self.num_classes = 10
        self.head_cfg = dict(
            type="SemiLinearClsHead",
            in_channels=self.in_channels,
            num_classes=self.num_classes,
        )

    @e2e_pytest_unit
    def test_build_semisl_cls_head(self):
        """Verifies that SemiSLClsHead builds."""
        head = build_head(self.head_cfg)
        assert isinstance(head, SemiLinearClsHead)

        head_cfg = dict(
            type="SemiNonLinearClsHead",
            in_channels=self.in_channels,
            num_classes=self.num_classes,
        )
        head = build_head(head_cfg)
        assert isinstance(head, SemiNonLinearClsHead)

    @e2e_pytest_unit
    def test_build_semisl_cls_head_type_error(self):
        """Verifies that SemiSLClsHead parameters check with TypeError."""
        with pytest.raises(TypeError):
            self.head_cfg["num_classes"] = [1]
            build_head(self.head_cfg)
        with pytest.raises(TypeError):
            self.head_cfg["in_channels"] = [1]
            build_head(self.head_cfg)
        with pytest.raises(TypeError):
            self.head_cfg["loss"] = [1]
            build_head(self.head_cfg)
        with pytest.raises(TypeError):
            self.head_cfg["topk"] = [1]
            build_head(self.head_cfg)
        with pytest.raises(TypeError):
            self.head_cfg["unlabeled_coef"] = [1]
            build_head(self.head_cfg)
        with pytest.raises(TypeError):
            self.head_cfg["min_threshold"] = [1]
            build_head(self.head_cfg)

    @e2e_pytest_unit
    def test_build_semisl_cls_head_value_error(self):
        """Verifies that SemiSLClsHead parameters check with ValueError."""
        with pytest.raises(ValueError):
            self.head_cfg["num_classes"] = 0
            build_head(self.head_cfg)
        with pytest.raises(ValueError):
            self.head_cfg["num_classes"] = -1
            build_head(self.head_cfg)
        with pytest.raises(ValueError):
            self.head_cfg["in_channels"] = 0
            build_head(self.head_cfg)
        with pytest.raises(ValueError):
            self.head_cfg["in_channels"] = -1
            build_head(self.head_cfg)

    @e2e_pytest_unit
    def test_forward(self, mocker):
        """Verifies that SemiSLClsHead forward function works."""
        head = build_head(self.head_cfg)
        labeled_batch_size = 16
        unlabeled_batch_size = 64

        dummy_gt = torch.randint(self.num_classes, (labeled_batch_size,))
        labeled = torch.rand(labeled_batch_size, self.in_channels)
        unlabeled_weak = torch.rand(unlabeled_batch_size, self.in_channels)
        unlabeled_strong = torch.rand(unlabeled_batch_size, self.in_channels)

        mocker.patch("torch.cuda.is_available", return_value=False)

        dummy_data = {
            "labeled": labeled,
            "unlabeled_weak": unlabeled_weak,
            "unlabeled_strong": unlabeled_strong,
        }
        head.classwise_acc = head.classwise_acc.cpu()
        loss = head.forward_train(dummy_data, dummy_gt)

        assert isinstance(loss, dict)
        assert len(loss) == 3
        assert isinstance(loss["accuracy"], dict)
        assert len(loss["accuracy"]) == 2

        # No Unlabeled Data
        dummy_feature = torch.rand(labeled_batch_size, self.in_channels)
        loss = head.forward_train(dummy_feature, dummy_gt)

        assert isinstance(loss, dict)
        assert len(loss) == 3
        assert isinstance(loss["accuracy"], dict)
        assert len(loss["accuracy"]) == 2

    @e2e_pytest_unit
    def test_simple_test(self):
        """Verifies that SemiSLClsHead simple_test function works."""
        head = build_head(self.head_cfg)
        dummy_feature = torch.rand(3, self.in_channels)
        features = head.simple_test(dummy_feature)
        assert len(features) == 3
        assert len(features[0]) == self.num_classes



"""NNCF-related utils."""

# Copyright (C) 2021-2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import torch


def is_checkpoint_nncf(path: str) -> bool:
    """Check if checkpoint is NNCF checkpoint.

    The function uses metadata stored in a checkpoint to check if the
    checkpoint was the result of training of NNCF-compressed model.
    """
    state = torch.load(path, map_location="cpu")
    is_nncf = bool(state.get("meta", {}).get("nncf_enable_compression")) or "nncf_metainfo" in state
    return is_nncf


def get_number_of_fakequantizers_in_xml(path_to_xml: str) -> int:
    """Return number of FakeQuantize layers.

    Return number of FakeQuantize layers in the model by parsing file without loading model.

    Args:
        path_to_xml (str): Path to xml file.

    Returns:
        int: Number of FakeQuantize layers.
    """
    num_fq = 0
    with open(path_to_xml, "r", encoding="UTF-8") as stream:
        for line in stream.readlines():
            if 'type="FakeQuantize"' in line:
                num_fq += 1
    return num_fq



import os
import warnings

import numpy as np

from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.color import Color
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.label_schema import LabelGroup, LabelGroupType, LabelSchemaEntity
from otx.api.entities.shapes.ellipse import Ellipse
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.task_environment import TaskEnvironment
from tests.test_helpers import generate_random_annotated_image

DEFAULT_SEG_TEMPLATE_DIR = os.path.join("otx/algorithms/segmentation/configs", "ocr_lite_hrnet_18_mod2")
DEFAULT_RECIPE_CONFIG_PATH = "otx/recipes/stages/segmentation/incremental.py"

labels_names = ("rectangle", "ellipse", "triangle")


def generate_otx_label_schema(labels_names=labels_names):
    label_domain = Domain.SEGMENTATION
    rgb = [int(i) for i in np.random.randint(0, 256, 3)]
    colors = [Color(*rgb) for _ in range(len(labels_names))]
    not_empty_labels = [
        LabelEntity(name=name, color=colors[i], domain=label_domain, id=i) for i, name in enumerate(labels_names)
    ]
    empty_label = LabelEntity(
        name="Empty label",
        color=Color(42, 43, 46),
        is_empty=True,
        domain=label_domain,
        id=len(not_empty_labels),
    )

    label_schema = LabelSchemaEntity()
    exclusive_group = LabelGroup(name="labels", labels=not_empty_labels, group_type=LabelGroupType.EXCLUSIVE)
    empty_group = LabelGroup(name="empty", labels=[empty_label], group_type=LabelGroupType.EMPTY_LABEL)
    label_schema.add_group(exclusive_group)
    label_schema.add_group(empty_group)
    return label_schema


def init_environment(params, model_template):
    labels_schema = generate_otx_label_schema()
    environment = TaskEnvironment(
        model=None,
        hyper_parameters=params,
        label_schema=labels_schema,
        model_template=model_template,
    )

    return environment


def generate_otx_dataset(number_of_images=5):
    items = []
    labels_schema = generate_otx_label_schema()
    labels_list = labels_schema.get_labels(False)
    for i in range(0, number_of_images):
        image_numpy, shapes = generate_random_annotated_image(
            image_width=640,
            image_height=480,
            labels=labels_list,
            max_shapes=20,
            min_size=50,
            max_size=100,
            random_seed=None,
        )
        # Convert all shapes to polygons
        out_shapes = []
        for shape in shapes:
            shape_labels = shape.get_labels(include_empty=True)

            in_shape = shape.shape
            if isinstance(in_shape, Rectangle):
                points = [
                    Point(in_shape.x1, in_shape.y1),
                    Point(in_shape.x2, in_shape.y1),
                    Point(in_shape.x2, in_shape.y2),
                    Point(in_shape.x1, in_shape.y2),
                ]
            elif isinstance(in_shape, Ellipse):
                points = [Point(x, y) for x, y in in_shape.get_evenly_distributed_ellipse_coordinates()]
            elif isinstance(in_shape, Polygon):
                points = in_shape.points

            out_shapes.append(Annotation(Polygon(points=points), labels=shape_labels))

        image = Image(data=image_numpy)
        annotation = AnnotationSceneEntity(kind=AnnotationSceneKind.ANNOTATION, annotations=out_shapes)
        items.append(DatasetItemEntity(media=image, annotation_scene=annotation))
    warnings.resetwarnings()

    dataset = DatasetEntity(items)

    return dataset



"""This file defines the ShapeEntity interface and the Shape abstract class."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc
import datetime
import warnings
from enum import IntEnum, auto
from typing import TYPE_CHECKING

from shapely.errors import PredicateError, TopologicalError
from shapely.geometry import Polygon as shapely_polygon

if TYPE_CHECKING:
    from otx.api.entities.shapes.rectangle import Rectangle


class GeometryException(ValueError):
    """Exception that is thrown if the geometry of a Shape is invalid."""


class ShapeType(IntEnum):
    """Shows which type of Shape is being used."""

    ELLIPSE = auto()
    RECTANGLE = auto()
    POLYGON = auto()


class ShapeEntity(metaclass=abc.ABCMeta):
    """This interface represents the annotation shapes on the media given by user annotations or system analysis.

    The shapes is a 2D geometric shape living in a normalized coordinate system (the values range from 0 to 1).
    """

    # pylint: disable=redefined-builtin
    def __init__(self, shape_type: ShapeType):
        self._type = shape_type

    @property
    def type(self) -> ShapeType:
        """Get the type of Shape that this Shape represents."""
        return self._type

    @abc.abstractmethod
    def get_area(self) -> float:
        """Get the area of the shape."""
        raise NotImplementedError

    @abc.abstractmethod
    def intersects(self, other: "Shape") -> bool:
        """Returns true if other intersects with shape, otherwise returns false.

        Args:
            other (Shape): Shape to compare with

        Returns:
            bool: true if other intersects with shape, otherwise returns false
        """
        raise NotImplementedError

    @abc.abstractmethod
    def contains_center(self, other: "ShapeEntity") -> bool:
        """Checks whether the center of the 'other' shape is located in the shape.

        Args:
            other (ShapeEntity): Shape to compare with

        Returns:
            bool: true if the center of the 'other' shape is located in the shape, otherwise returns false
        """
        raise NotImplementedError

    @abc.abstractmethod
    def normalize_wrt_roi_shape(self, roi_shape: "Rectangle") -> "Shape":
        """The inverse of denormalize_wrt_roi_shape.

        Transforming shape from the `roi` coordinate system to the normalized coordinate system.
        This is used when the tasks want to save the analysis results.

        For example in Detection -> Segmentation pipeline, the analysis results of segmentation
        needs to be normalized to the roi (bounding boxes) coming from the detection.

        Args:
            roi_shape (Rectangle): Shape of the roi.

        Returns:
            Shape: Shape in the normalized coordinate system.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def denormalize_wrt_roi_shape(self, roi_shape: "Rectangle") -> "ShapeEntity":
        """The inverse of normalize_wrt_roi_shape.

        Transforming shape from the normalized coordinate system to the `roi` coordinate system.
        This is used to pull ground truth during training process of the tasks.
        Examples given in the Shape implementations.

        Args:
            roi_shape (Rectangle): Shape of the roi.

        Returns:
            ShapeEntity: Shape in the `roi` coordinate system.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _as_shapely_polygon(self) -> shapely_polygon:
        """Convert shape to a shapely polygon.

        Shapely polygons are within the SDK used to calculate the intersection between Shapes.
        It is also used in the SDK to find shapes that are visible within a given ROI.

        Returns:
            shapely_polygon: Shapely polygon representation of the shape.
        """
        raise NotImplementedError


class Shape(ShapeEntity):
    """Base class for Shape entities."""

    # pylint: disable=redefined-builtin, too-many-arguments; Requires refactor
    def __init__(self, shape_type: ShapeType, modification_date: datetime.datetime):
        super().__init__(shape_type=shape_type)
        self.modification_date = modification_date

    def __repr__(self):
        """Returns the date of the last modification of the shape."""
        return f"Shape with modification date:('{self.modification_date}')"

    def get_area(self) -> float:
        """Get the area of the shape."""
        raise NotImplementedError

    # pylint: disable=protected-access
    def intersects(self, other: "Shape") -> bool:
        """Returns True, if other intersects with shape, otherwise returns False."""
        polygon_roi = self._as_shapely_polygon()
        polygon_shape = other._as_shapely_polygon()
        try:
            return polygon_roi.intersects(polygon_shape)
        except (PredicateError, TopologicalError) as exception:
            raise GeometryException(
                f"The intersection between the shapes {self} and {other} could not be computed: " f"{exception}."
            ) from exception

    # pylint: disable=protected-access
    def contains_center(self, other: "ShapeEntity") -> bool:
        """Checks whether the center of the 'other' shape is located in the shape.

        Args:
            other (ShapeEntity): Shape to compare with.

        Returns:
            bool: Boolean that indicates whether the center of the other shape is located in the shape
        """
        polygon_roi = self._as_shapely_polygon()
        polygon_shape = other._as_shapely_polygon()
        return polygon_roi.contains(polygon_shape.centroid)

    def _validate_coordinates(self, x: float, y: float) -> bool:
        """Check if coordinate is valid.

        Checks whether the values for a given x,y coordinate pair lie within the range of (0,1) that is expected for
        the normalized coordinate system. Issues a warning if the coordinates are out of bounds.

        Args:
            x (float): x-coordinate to validate
            y (float): y-coordinate to validate

        Returns:
            bool: ``True`` if coordinates are within expected range, ``False`` otherwise
        """
        if not ((0.0 <= x <= 1.0) and (0.0 <= y <= 1.0)):
            warnings.warn(
                f"{type(self).__name__} coordinates (x={x}, y={y}) are out of bounds, a normalized "
                f"coordinate system is assumed. All coordinates are expected to be in range (0,1).",
                UserWarning,
            )
            return False
        return True

    def __hash__(self):
        """Returns the hash of shape."""
        return hash(str(self))



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import glob
import os.path as osp
import time

from mmcls import __version__
from mmcls.apis import train_model
from mmcls.datasets import build_dataloader, build_dataset
from mmcls.utils import collect_env
from torch import nn

from otx.mpa.registry import STAGES
from otx.mpa.utils.logger import get_logger

from .stage import ClsStage

logger = get_logger()


@STAGES.register_module()
class ClsTrainer(ClsStage):
    def run(self, model_cfg, model_ckpt, data_cfg, **kwargs):  # noqa: C901
        """Run training stage for classification

        - Configuration
        - Environment setup
        - Run training via MMClassification -> MMCV
        """
        self._init_logger()
        mode = kwargs.get("mode", "train")
        if mode not in self.mode:
            logger.warning(f"Supported modes are {self.mode} but '{mode}' is given.")
            return {}

        cfg = self.configure(model_cfg, model_ckpt, data_cfg, training=True, **kwargs)
        logger.info("train!")

        timestamp = time.strftime("%Y%m%d_%H%M%S", time.localtime())

        # Environment
        logger.info(f"cfg.gpu_ids = {cfg.gpu_ids}, distributed = {self.distributed}")
        env_info_dict = collect_env()
        env_info = "\n".join([(f"{k}: {v}") for k, v in env_info_dict.items()])
        dash_line = "-" * 60 + "\n"
        logger.info("Environment info:\n" + dash_line + env_info + "\n" + dash_line)

        # Data
        datasets = [build_dataset(cfg.data.train)]

        # Metadata
        meta = dict()
        meta["env_info"] = env_info
        # meta['config'] = cfg.pretty_text
        meta["seed"] = cfg.seed

        repr_ds = datasets[0]

        if cfg.checkpoint_config is not None:
            cfg.checkpoint_config.meta = dict(mmcls_version=__version__)
            if hasattr(repr_ds, "tasks"):
                cfg.checkpoint_config.meta["tasks"] = repr_ds.tasks
            else:
                cfg.checkpoint_config.meta["CLASSES"] = repr_ds.CLASSES
            if "task_adapt" in cfg:
                if hasattr(self, "model_tasks"):  # for incremnetal learning
                    cfg.checkpoint_config.meta.update({"tasks": self.model_tasks})
                    # instead of update(self.old_tasks), update using "self.model_tasks"
                else:
                    cfg.checkpoint_config.meta.update({"CLASSES": self.model_classes})

        self.configure_samples_per_gpu(cfg, "train", self.distributed)
        self.configure_fp16_optimizer(cfg, self.distributed)

        # Model
        model_builder = kwargs.get("model_builder", None)
        model = self.build_model(cfg, model_builder)
        model.train()

        if self.distributed:
            self._modify_cfg_for_distributed(model, cfg)

        self.configure_compat_cfg(cfg)

        # Save config
        # cfg.dump(osp.join(cfg.work_dir, 'config.py'))
        # logger.info(f'Config:\n{cfg.pretty_text}')

        # register custom eval hooks
        validate = True if cfg.data.get("val", None) else False
        if validate:
            val_dataset = build_dataset(cfg.data.val, dict(test_mode=True))
            val_loader_cfg = {
                # cfg.gpus will be ignored if distributed
                "num_gpus": len(cfg.gpu_ids),
                "dist": self.distributed,
                "round_up": True,
                "seed": cfg.seed,
                "shuffle": False,  # Not shuffle by default
                "sampler_cfg": None,  # Not use sampler by default
                **cfg.data.get("val_dataloader", {}),
            }
            val_dataloader = build_dataloader(val_dataset, **val_loader_cfg)
            eval_cfg = cfg.get("evaluation", {})
            eval_cfg["by_epoch"] = cfg.runner["type"] != "IterBasedRunner"
            cfg.custom_hooks.append(
                dict(
                    type="DistCustomEvalHook" if self.distributed else "CustomEvalHook",
                    dataloader=val_dataloader,
                    priority="ABOVE_NORMAL",
                    **eval_cfg,
                )
            )

        train_model(
            model,
            datasets,
            cfg,
            distributed=self.distributed,
            validate=False,
            timestamp=timestamp,
            meta=meta,
        )

        # Save outputs
        output_ckpt_path = osp.join(cfg.work_dir, "latest.pth")
        best_ckpt_path = glob.glob(osp.join(cfg.work_dir, "best_*.pth"))
        if len(best_ckpt_path) > 0:
            output_ckpt_path = best_ckpt_path[0]
        return dict(
            final_ckpt=output_ckpt_path,
        )

    def _modify_cfg_for_distributed(self, model, cfg):
        nn.SyncBatchNorm.convert_sync_batchnorm(model)

        if cfg.dist_params.get("linear_scale_lr", False):
            new_lr = len(cfg.gpu_ids) * cfg.optimizer.lr
            logger.info(
                f"enabled linear scaling rule to the learning rate. \
                changed LR from {cfg.optimizer.lr} to {new_lr}"
            )
            cfg.optimizer.lr = new_lr



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest
import torch

from otx.mpa.modules.models.losses.asymmetric_angular_loss_with_ignore import (
    AsymmetricAngularLossWithIgnore,
)
from otx.mpa.modules.models.losses.asymmetric_loss_with_ignore import (
    AsymmetricLossWithIgnore,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestAsymmetricLoss:
    @pytest.fixture(autouse=True)
    def loss_type(self) -> None:
        return AsymmetricLossWithIgnore

    @pytest.fixture(autouse=True)
    def setup(self, loss_type) -> None:
        self.num_classes = 2
        self.default_gt = torch.zeros((2, self.num_classes))
        self.default_input = torch.zeros((2, self.num_classes))
        self.default_input[0, 0] = 1
        self.default_gt[0, 0] = 1
        self.default_loss = loss_type(reduction="mean")

    @e2e_pytest_unit
    def test_forward(self) -> None:
        result_c = self.default_loss(self.default_input, self.default_gt)
        self.default_input[0, 1] = 1
        result_w = self.default_loss(self.default_input, self.default_gt)
        assert result_c < result_w

    @e2e_pytest_unit
    def test_weight(self, loss_type) -> None:
        result = self.default_loss(self.default_input, self.default_gt)
        loss_w = loss_type(loss_weight=2, reduction="mean")
        result_w = loss_w(self.default_input, self.default_gt)
        assert result_w > result

    @e2e_pytest_unit
    def test_reduction(self, loss_type) -> None:
        result = self.default_loss(self.default_input, self.default_gt)
        loss_s = loss_type(reduction="sum")
        result_s = loss_s(self.default_input, self.default_gt)
        assert result_s > result

    @e2e_pytest_unit
    def test_gamma_neg(self, loss_type) -> None:
        result = self.default_loss(self.default_input, self.default_gt)
        loss_s = loss_type(gamma_neg=0.0, reduction="mean")
        result_s = loss_s(self.default_input, self.default_gt)
        assert result_s > result

    @e2e_pytest_unit
    def test_gamma_pos(self, loss_type) -> None:
        result = self.default_loss(self.default_input, self.default_gt)
        loss_s = loss_type(gamma_pos=1.0, reduction="mean")
        result_s = loss_s(self.default_input, self.default_gt)
        assert result_s < result


class TestAsymmetricAngularLoss(TestAsymmetricLoss):
    @pytest.fixture(autouse=True)
    def loss_type(self) -> None:
        return AsymmetricAngularLossWithIgnore



"""Task Initialization of OTX Classification."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import otx.algorithms.classification.adapters.mmcls as MPAData
import otx.algorithms.common.adapters.mmcv.models as OTXBackbones
import otx.mpa.cls as MPAClassification

from .inference import ClassificationInferenceTask
from .nncf import ClassificationNNCFTask
from .openvino import ClassificationOpenVINOTask
from .train import ClassificationTrainTask

__all__ = [
    "MPAClassification",
    "OTXBackbones",
    "MPAData",
    "ClassificationOpenVINOTask",
    "ClassificationInferenceTask",
    "ClassificationTrainTask",
    "ClassificationNNCFTask",
]



"""Data Pipeline of Self-SL model for Classification Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

__img_norm_cfg = dict(mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_rgb=True)
__img_size = 224

__train_pipeline_v0 = [
    dict(type="RandomResizedCrop", size=__img_size),
    dict(type="RandomFlip"),
    dict(
        type="RandomAppliedTrans",
        transforms=[
            dict(type="OTXColorJitter", brightness=0.4, contrast=0.4, saturation=0.2, hue=0.1),
        ],
        p=0.8,
    ),
    dict(type="RandomGrayscale", gray_prob=0.2),
    dict(type="GaussianBlur", sigma_min=0.1, sigma_max=2.0),
    dict(type="Normalize", **__img_norm_cfg),
    dict(type="ImageToTensor", keys=["img"]),
    dict(type="Collect", keys=["img"]),
]
__train_pipeline_v1 = [
    dict(type="RandomResizedCrop", size=__img_size),
    dict(type="RandomFlip"),
    dict(
        type="RandomAppliedTrans",
        transforms=[
            dict(type="OTXColorJitter", brightness=0.4, contrast=0.4, saturation=0.2, hue=0.1),
        ],
        p=0.8,
    ),
    dict(type="RandomGrayscale", gray_prob=0.2),
    dict(type="RandomAppliedTrans", transforms=[dict(type="GaussianBlur", sigma_min=0.1, sigma_max=2.0)], p=0.1),
    dict(type="RandomAppliedTrans", transforms=[dict(type="Solarize", thr=128)], p=0.2),
    dict(type="Normalize", **__img_norm_cfg),
    dict(type="ImageToTensor", keys=["img"]),
    dict(type="Collect", keys=["img"]),
]

data = dict(
    train=dict(pipeline=dict(view0=__train_pipeline_v0, view1=__train_pipeline_v1)),
)



"""OTX Adapters for action recognition backbones - mmaction2."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from mmaction.models.backbones.x3d import X3D
from mmdet.models import BACKBONES as MMDET_BACKBONES


def register_action_backbones():
    """Register action backbone to mmdetection backbones."""
    MMDET_BACKBONES.register_module()(X3D)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
from mmcv.runner import force_fp32
from mmdet.core import bbox2roi, multi_apply
from mmdet.models.builder import HEADS, build_head, build_roi_extractor
from mmdet.models.losses import accuracy
from mmdet.models.roi_heads.bbox_heads.convfc_bbox_head import Shared2FCBBoxHead
from mmdet.models.roi_heads.standard_roi_head import StandardRoIHead

from otx.mpa.modules.models.heads.cross_dataset_detector_head import (
    CrossDatasetDetectorHead,
)
from otx.mpa.modules.models.losses.cross_focal_loss import CrossSigmoidFocalLoss


@HEADS.register_module()
class CustomRoIHead(StandardRoIHead):
    def init_bbox_head(self, bbox_roi_extractor, bbox_head):
        """Initialize ``bbox_head``"""
        self.bbox_roi_extractor = build_roi_extractor(bbox_roi_extractor)
        if bbox_head.type == "Shared2FCBBoxHead":
            bbox_head.type = "CustomConvFCBBoxHead"
        self.bbox_head = build_head(bbox_head)

    def _bbox_forward_train(self, x, sampling_results, gt_bboxes, gt_labels, img_metas):
        """Run forward function and calculate loss for box head in training."""
        rois = bbox2roi([res.bboxes for res in sampling_results])
        bbox_results = self._bbox_forward(x, rois)

        labels, label_weights, bbox_targets, bbox_weights, valid_label_mask = self.bbox_head.get_targets(
            sampling_results, gt_bboxes, gt_labels, img_metas, self.train_cfg
        )
        loss_bbox = self.bbox_head.loss(
            bbox_results["cls_score"],
            bbox_results["bbox_pred"],
            rois,
            labels,
            label_weights,
            bbox_targets,
            bbox_weights,
            valid_label_mask=valid_label_mask,
        )
        bbox_results.update(loss_bbox=loss_bbox)
        return bbox_results


@HEADS.register_module()
class CustomConvFCBBoxHead(Shared2FCBBoxHead, CrossDatasetDetectorHead):
    def get_targets(self, sampling_results, gt_bboxes, gt_labels, img_metas, rcnn_train_cfg, concat=True):
        """Calculate the ground truth for all samples in a batch according to
        the sampling_results.

        Almost the same as the implementation in bbox_head, we passed
        additional parameters pos_inds_list and neg_inds_list to
        `_get_target_single` function.

        Args:
            sampling_results (List[obj:SamplingResults]): Assign results of
                all images in a batch after sampling.
            gt_bboxes (list[Tensor]): Gt_bboxes of all images in a batch,
                each tensor has shape (num_gt, 4),  the last dimension 4
                represents [tl_x, tl_y, br_x, br_y].
            gt_labels (list[Tensor]): Gt_labels of all images in a batch,
                each tensor has shape (num_gt,).
            rcnn_train_cfg (obj:ConfigDict): `train_cfg` of RCNN.
            concat (bool): Whether to concatenate the results of all
                the images in a single batch.

        Returns:
            Tuple[Tensor]: Ground truth for proposals in a single image.
            Containing the following list of Tensors:

                - labels (list[Tensor],Tensor): Gt_labels for all
                  proposals in a batch, each tensor in list has
                  shape (num_proposals,) when `concat=False`, otherwise
                  just a single tensor has shape (num_all_proposals,).
                - label_weights (list[Tensor]): Labels_weights for
                  all proposals in a batch, each tensor in list has
                  shape (num_proposals,) when `concat=False`, otherwise
                  just a single tensor has shape (num_all_proposals,).
                - bbox_targets (list[Tensor],Tensor): Regression target
                  for all proposals in a batch, each tensor in list
                  has shape (num_proposals, 4) when `concat=False`,
                  otherwise just a single tensor has shape
                  (num_all_proposals, 4), the last dimension 4 represents
                  [tl_x, tl_y, br_x, br_y].
                - bbox_weights (list[tensor],Tensor): Regression weights for
                  all proposals in a batch, each tensor in list has shape
                  (num_proposals, 4) when `concat=False`, otherwise just a
                  single tensor has shape (num_all_proposals, 4).
        """
        pos_bboxes_list = [res.pos_bboxes for res in sampling_results]
        neg_bboxes_list = [res.neg_bboxes for res in sampling_results]
        pos_gt_bboxes_list = [res.pos_gt_bboxes for res in sampling_results]
        pos_gt_labels_list = [res.pos_gt_labels for res in sampling_results]
        labels, label_weights, bbox_targets, bbox_weights = multi_apply(
            self._get_target_single,
            pos_bboxes_list,
            neg_bboxes_list,
            pos_gt_bboxes_list,
            pos_gt_labels_list,
            cfg=rcnn_train_cfg,
        )
        valid_label_mask = self.get_valid_label_mask(img_metas=img_metas, all_labels=labels, use_bg=True)
        valid_label_mask = [i.to(gt_bboxes[0].device) for i in valid_label_mask]

        if concat:
            labels = torch.cat(labels, 0)
            label_weights = torch.cat(label_weights, 0)
            bbox_targets = torch.cat(bbox_targets, 0)
            bbox_weights = torch.cat(bbox_weights, 0)
            valid_label_mask = torch.cat(valid_label_mask, 0)
        return labels, label_weights, bbox_targets, bbox_weights, valid_label_mask

    @force_fp32(apply_to=("cls_score", "bbox_pred"))
    def loss(
        self,
        cls_score,
        bbox_pred,
        rois,
        labels,
        label_weights,
        bbox_targets,
        bbox_weights,
        reduction_override=None,
        valid_label_mask=None,
    ):
        losses = dict()
        if cls_score is not None and cls_score.numel() > 0:
            avg_factor = max(torch.sum(label_weights > 0).float().item(), 1.0)
            if isinstance(self.loss_cls, CrossSigmoidFocalLoss):
                losses["loss_cls"] = self.loss_cls(
                    cls_score,
                    labels,
                    label_weights,
                    avg_factor=avg_factor,
                    reduction_override=reduction_override,
                    use_bg=True,
                    valid_label_mask=valid_label_mask,
                )
            else:
                losses["loss_cls"] = self.loss_cls(
                    cls_score, labels, label_weights, avg_factor=avg_factor, reduction_override=reduction_override
                )
            losses["acc"] = accuracy(cls_score, labels)
        if bbox_pred is not None:
            bg_class_ind = self.num_classes
            # 0~self.num_classes-1 are FG, self.num_classes is BG
            pos_inds = (labels >= 0) & (labels < bg_class_ind)
            # do not perform bounding box regression for BG anymore.
            if pos_inds.any():
                if self.reg_decoded_bbox:
                    # When the regression loss (e.g. `IouLoss`,
                    # `GIouLoss`, `DIouLoss`) is applied directly on
                    # the decoded bounding boxes, it decodes the
                    # already encoded coordinates to absolute format.
                    bbox_pred = self.bbox_coder.decode(rois[:, 1:], bbox_pred)
                if self.reg_class_agnostic:
                    pos_bbox_pred = bbox_pred.view(bbox_pred.size(0), 4)[pos_inds.type(torch.bool)]
                else:
                    pos_bbox_pred = bbox_pred.view(bbox_pred.size(0), -1, 4)[
                        pos_inds.type(torch.bool), labels[pos_inds.type(torch.bool)]
                    ]
                losses["loss_bbox"] = self.loss_bbox(
                    pos_bbox_pred,
                    bbox_targets[pos_inds.type(torch.bool)],
                    bbox_weights[pos_inds.type(torch.bool)],
                    avg_factor=bbox_targets.size(0),
                    reduction_override=reduction_override,
                )
            else:
                losses["loss_bbox"] = bbox_pred[pos_inds].sum()
        return losses



"""This file defines the Subset enum for use in datasets."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from enum import Enum


class Subset(Enum):
    """Describes the Subset a DatasetItem is assigned to."""

    NONE = 0
    TRAINING = 1
    VALIDATION = 2
    TESTING = 3
    UNLABELED = 4
    PSEUDOLABELED = 5
    UNASSIGNED = 6

    def __str__(self):
        """Returns name of subset."""
        return str(self.name)

    def __repr__(self):
        """Returns name of subset."""
        return f"Subset.{self.name}"



from otx.mpa.version import __version__, get_version
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@e2e_pytest_unit
def test_version():
    return_value = get_version()

    assert return_value == __version__



"""NNCF task related hooks."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcv.runner.hooks.hook import HOOKS, Hook


@HOOKS.register_module()
class CompressionHook(Hook):
    """CompressionHook."""

    COMPRESSION_STATE_FILE_NAME = "meta_state.pth"

    def __init__(self, compression_ctrl=None):
        self.compression_ctrl = compression_ctrl

    def after_train_iter(self, runner):
        """Called after train iter."""
        self.compression_ctrl.scheduler.step()

    def after_train_epoch(self, runner):
        """Called after train epoch."""
        self.compression_ctrl.scheduler.epoch_step()
        if runner.rank == 0:
            runner.logger.info(self.compression_ctrl.statistics().to_str())

    def before_run(self, runner):
        """Called before run."""
        runner.compression_ctrl = self.compression_ctrl
        if runner.rank == 0:
            runner.logger.info(self.compression_ctrl.statistics().to_str())



"""Data Pipeline for SupCon model of Segmentation Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

__img_norm_cfg = dict(mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_rgb=True)
__img_scale = (544, 544)
__resize_target_size = (512, 512)

__train_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(type="LoadAnnotations"),
    dict(
        type="TwoCropTransform",
        view0=[
            dict(type="NDArrayToPILImage", keys=["img"]),
            dict(type="RandomResizedCrop", size=__resize_target_size),
            dict(type="RandomColorJitter", brightness=0.4, contrast=0.4, saturation=0.2, hue=0.1, p=0.8),
            dict(type="RandomGrayscale", p=0.2),
            dict(type="RandomGaussianBlur", kernel_size=23, p=1.0),
            dict(type="PILImageToNDArray", keys=["img"]),
            dict(type="RandomFlip", prob=0.5, direction="horizontal"),
            dict(type="Normalize", **__img_norm_cfg),
        ],
        view1=[
            dict(type="NDArrayToPILImage", keys=["img"]),
            dict(type="RandomResizedCrop", size=__resize_target_size),
            dict(type="RandomColorJitter", brightness=0.4, contrast=0.4, saturation=0.2, hue=0.1, p=0.8),
            dict(type="RandomGrayscale", p=0.2),
            dict(type="RandomGaussianBlur", kernel_size=23, p=0.1),
            dict(type="PILImageToNDArray", keys=["img"]),
            dict(type="RandomSolarization", threshold=128, p=0.2),
            dict(type="RandomFlip", prob=0.5, direction="horizontal"),
            dict(type="Normalize", **__img_norm_cfg),
        ],
    ),
    dict(type="DefaultFormatBundle"),
    dict(type="Collect", keys=["img", "gt_semantic_seg"]),
]

__test_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(
        type="MultiScaleFlipAug",
        img_scale=__img_scale,
        flip=False,
        transforms=[
            dict(type="Resize", keep_ratio=False),
            dict(type="RandomFlip"),
            dict(type="Normalize", **__img_norm_cfg),
            dict(type="ImageToTensor", keys=["img"]),
            dict(type="Collect", keys=["img"]),
        ],
    ),
]

# TODO (Sungchul, Soobee) : Remove Repeatdataset in data config
# when otx/algorithms/segmentation/configs/base/data/data_pipeline.py is updated.
data = dict(
    train=dict(dataset=dict(pipeline=__train_pipeline)),
    val=dict(pipeline=__test_pipeline),
    test=dict(pipeline=__test_pipeline),
)



_base_ = [
    "../_base_/data/seg_semisl.py",
    "../_base_/models/segmentors/segmentor.py",
    "./train.py",
]

optimizer = dict(_delete_=True, type="Adam", lr=1e-3, eps=1e-08, weight_decay=0.0)

optimizer_config = dict(_delete_=True, grad_clip=dict(max_norm=40, norm_type=2))

custom_hooks = [
    dict(
        type="DualModelEMAHook",
        momentum=0.99,
        start_epoch=1,
        src_model_name="model_s",
        dst_model_name="model_t",
    ),
]

log_config = dict(
    interval=1,
    hooks=[
        dict(type="TextLoggerHook", by_epoch=True),
    ],
)

dist_params = dict(backend="nccl", linear_scale_lr=False)

runner = dict(type="EpochRunnerWithCancel", max_epochs=300)

checkpoint_config = dict(
    by_epoch=True,
    interval=1,
)

evaluation = dict(interval=1, metric=["mDice", "mIoU"], show_log=True)

task_adapt = dict(
    op="REPLACE",
)
ignore = True


find_unused_parameters = False
seed = 42



# Copyright (c) 2019 MendelXu
# SPDX-License-Identifier: Apache-2.0
#
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
import torch.nn as nn
import torch.nn.functional as F
from mmcv.cnn import ConvModule

from .psp_layer import PSPModule


class AsymmetricPositionAttentionModule(nn.Module):
    """Reference: https://github.com/MendelXu/ANN"""

    def __init__(
        self,
        in_channels,
        key_channels,
        value_channels=None,
        psp_size=(1, 3, 6, 8),
        conv_cfg=None,
        norm_cfg=dict(type="BN"),
    ):
        super().__init__()

        self.in_channels = in_channels
        self.key_channels = key_channels
        self.value_channels = value_channels if value_channels is not None else in_channels
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg

        self.query_key = ConvModule(
            in_channels=self.in_channels,
            out_channels=self.key_channels,
            kernel_size=1,
            stride=1,
            padding=0,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=dict(type="ReLU"),
        )
        self.key_psp = PSPModule(psp_size, method="max")

        self.value = ConvModule(
            in_channels=self.in_channels,
            out_channels=self.value_channels,
            kernel_size=1,
            stride=1,
            padding=0,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=dict(type="ReLU"),
        )
        self.value_psp = PSPModule(psp_size, method="max")

        self.out_conv = ConvModule(
            in_channels=self.value_channels,
            out_channels=self.in_channels,
            kernel_size=1,
            stride=1,
            padding=0,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=None,
        )

    def forward(self, x):
        batch_size, h, w = x.size(0), x.size(2), x.size(3)

        query_key = self.query_key(x)

        key = self.key_psp(query_key)
        value = self.value_psp(self.value(x)).permute(0, 2, 1)
        query = query_key.view(batch_size, self.key_channels, -1).permute(0, 2, 1)

        similarity_scores = torch.matmul(query, key)
        similarity_scores = (self.key_channels**-0.5) * similarity_scores
        similarity_scores = F.softmax(similarity_scores, dim=-1)

        y = torch.matmul(similarity_scores, value)
        y = y.permute(0, 2, 1).contiguous()
        y = y.view(batch_size, self.value_channels, *x.size()[2:])
        y = self.out_conv(y)

        out = x + y

        return out



"""Builder Class for training template.

For user's various use cases and convenient CLI,
It is an internal Builder class used in the otx build command
that enables the configuration of the basic workspace of OTX
and supports the replacement of the backbone of the model.
"""
# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import inspect
from pathlib import Path
from typing import Any, Dict, Union

import mmcv
import torch
from mmcv.utils import Registry, build_from_cfg
from torch import nn

from otx.api.entities.model_template import TaskType
from otx.cli.utils.importing import (
    get_backbone_list,
    get_backbone_registry,
    get_module_args,
)
from otx.mpa.utils.config_utils import MPAConfig

# pylint: disable=too-many-locals, too-many-statements, too-many-branches


def get_backbone_out_channels(backbone: nn.Module):
    """Get output channels of backbone using fake data."""
    out_channels = []
    input_size = backbone.input_size if hasattr(backbone, "input_size") else 64
    fake_data = torch.rand(2, 3, input_size, input_size)
    outputs = backbone(fake_data)
    for out in outputs:
        out_channels.append(out.shape[1])
    return out_channels


def update_backbone_args(backbone_config: dict, registry: Registry, backend: str):
    """Update Backbone required arguments.

    This function checks the init parameters of the corresponding backbone function (or class)
    and identifies the required arguments.
    Also, it distinguishes the argment needed for the build to add convenience to the user.
    """
    backbone_module = registry.get(backbone_config["type"])
    if not backbone_module:
        raise ValueError(f"{backbone_config['type']} is not supported backbone")
    required_args, default_args = get_module_args(backbone_module)
    for arg_key, default_value in default_args.items():
        if arg_key not in backbone_config:
            backbone_config[arg_key] = default_value

    missing_args = []
    for arg in required_args:
        if arg not in backbone_config:
            missing_args.append(arg)
    if len(missing_args) > 0:
        print(
            f"[*] {backbone_config['type']} requires the argument : {missing_args}"
            f"\n[*] Please refer to {inspect.getfile(backbone_module)}"
        )
    if "out_indices" in backbone_config:
        backbone_config["use_out_indices"] = True
    else:
        backbone_config["use_out_indices"] = False

    updated_missing_args = []
    backbone_type = backbone_config["type"]
    backbone_list = get_backbone_list(backend)
    if backbone_type not in backbone_list:
        return missing_args
    backbone_data = backbone_list[backbone_type]
    # Patch missing_args
    for arg in missing_args:
        if "options" in backbone_data and arg in backbone_data["options"]:
            backbone_config[arg] = backbone_data["options"][arg][0]
            print(
                f"[*] '{arg}' can choose between: {backbone_data['options'][arg]}"
                f"\n[*] '{arg}' default value: {backbone_config[arg]}"
            )
        else:
            backbone_config[arg] = "!!!!!!!!!!!INPUT_HERE!!!!!!!!!!!"
            updated_missing_args.append(arg)
    return updated_missing_args


def update_channels(model_config: MPAConfig, out_channels: Any):
    """Update in_channel of head or neck."""
    if hasattr(model_config.model, "neck"):
        if model_config.model.neck.type == "GlobalAveragePooling":
            model_config.model.neck.pop("in_channels", None)
        else:
            print(f"\tUpdate model.neck.in_channels: {out_channels}")
            model_config.model.neck.in_channels = out_channels

    elif hasattr(model_config.model, "decode_head"):
        head_in_index = model_config.model.decode_head.get("in_index", None)
        if head_in_index and len(out_channels) != len(head_in_index):
            updated_in_index = list(range(len(out_channels)))
            print(f"\tUpdate model.decode_head.in_index: {updated_in_index}")
            model_config.model.decode_head.in_index = updated_in_index
        print(f"\tUpdate model.decode_head.in_channels: {out_channels}")
        model_config.model.decode_head.in_channels = out_channels

    elif hasattr(model_config.model, "head"):
        print(f"\tUpdate model.head.in_channels: {out_channels}")
        model_config.model.head.in_channels = out_channels
    else:
        raise NotImplementedError("This architecture currently does not support public backbone.")


class Builder:
    """Class that implements a model templates registry."""

    def build_backbone_config(self, backbone_type: str, output_path: Union[Path, str]):
        """Build Backbone configs from backbone type.

        This is a function that makes the configuration
        of the usable backbone found by the user through otx find.
        backbone_type: The type of backbone want to get - {backend.backbone_type} (str)
        output_path: new backbone configuration file output path (Union[Path, str])
        """
        print(f"[*] Backbone Config: {backbone_type}")
        output_path = output_path if isinstance(output_path, Path) else Path(output_path)

        backend, backbone_class = Registry.split_scope_key(backbone_type)
        backbone_config: Dict[str, Any] = dict(type=backbone_type)
        if backbone_class == "MMOVBackbone":
            backend = f"omz.{backend}"
            backbone_config["verify_shape"] = False
        backbone_registry, _ = get_backbone_registry(backend)
        missing_args = update_backbone_args(backbone_config, backbone_registry, backend)
        if str(output_path).endswith((".yml", ".yaml", ".json")):
            mmcv.dump({"backbone": backbone_config}, str(output_path.absolute()))
            print(f"[*] Save backbone configuration: {str(output_path.absolute())}")
        else:
            raise ValueError("The backbone config support file format is as follows: (.yml, .yaml, .json)")
        return missing_args

    def merge_backbone(
        self,
        model_config_path: Union[Path, str],
        backbone_config_path: Union[Path, str],
        output_path: Union[Path, str] = None,
    ):
        """Build model & update backbone configs.

        This is a function that updates the existing model to be able to build
        through the backbone configuration file or backbone type.
        model_config_path: model configuration file path (Union[Path, str])
        backbone_config_path: backbone configuration file path (Union[Path, str])
        output_path: new model.py output path (Union[Path, str])
        """
        print(f"[*] Update {model_config_path} with {backbone_config_path}")
        model_config_path = model_config_path if isinstance(model_config_path, Path) else Path(model_config_path)
        backbone_config_path = (
            backbone_config_path if isinstance(backbone_config_path, Path) else Path(backbone_config_path)
        )

        # Get Model config from model config file
        if model_config_path.exists():
            model_config = MPAConfig.fromfile(str(model_config_path))
            print(f"\tTarget Model: {model_config.model.type}")
        else:
            raise ValueError(f"[*] The model is not properly defined or not found: {model_config_path}")

        # Get Backbone config from config file
        if backbone_config_path.exists():
            backbone_config = mmcv.load(str(backbone_config_path))
        else:
            raise ValueError(f"[*] The backbone is not found: {str(backbone_config_path)}")

        if "backbone" in backbone_config:
            backbone_config = backbone_config["backbone"]

        # Get Backbone configuration
        backend, backbone_class = Registry.split_scope_key(backbone_config["type"])
        backend = f"omz.{backend}" if backbone_class == "MMOVBackbone" else backend
        print(f"\tTarget Backbone: {backbone_config['type']}")
        otx_registry, custom_imports = get_backbone_registry(backend)

        # Update out_indices of backbone
        if backbone_config["use_out_indices"]:
            model_in_indices = []
            if "backbone" in model_config.model:
                model_in_indices = model_config.model.backbone.get("out_indices", [])
            backbone_out_indices = backbone_config.get("out_indices", None)
            if not backbone_out_indices and model_in_indices:
                # Check out_indices vs num_stage
                backbone_config["out_indices"] = model_in_indices
        backbone_config.pop("use_out_indices", None)

        # Build Backbone
        backbone = build_from_cfg(backbone_config, otx_registry, None)
        if model_config.model.get("task", None) == str(TaskType.CLASSIFICATION).lower():
            # Update model layer's in/out configuration in ClsStage.configure_model
            out_channels = -1
            if hasattr(model_config.model, "head"):
                model_config.model.head.in_channels = -1
        else:
            # Need to update in/out channel configuration here
            out_channels = get_backbone_out_channels(backbone)
        update_channels(model_config, out_channels)

        # Update Model Configuration
        if backend in ("torchvision"):
            backbone_config["init_cfg"] = {"Pretrained": True}
        print(f"\tBackbone config: {backbone_config}")
        model_config.model.backbone = backbone_config
        model_config.load_from = None

        if custom_imports:
            model_config["custom_imports"] = dict(imports=custom_imports, allow_failed_imports=False)

        # Dump or create model config file
        if output_path is None:
            output_path = model_config_path
        model_config.dump(str(output_path))
        print(f"[*] Save model configuration: {str(output_path)}")



"""NNCFNetwork patch util functions for mmcv models."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from contextlib import contextmanager
from functools import partial

from otx.algorithms.common.adapters.nncf.utils import nncf_trace, no_nncf_trace


@contextmanager
def nncf_trace_context(self, img_metas, nncf_compress_postprocessing=True):
    """A context manager for nncf graph tracing."""

    # onnx_export in mmdet head has a bug on GPU
    # it must be on CPU
    device_backup = next(self.parameters()).device  # pylint: disable=stop-iteration-return
    self = self.to("cpu")

    if nncf_compress_postprocessing:
        self.forward = partial(self.forward, img_metas=img_metas, return_loss=False)
    else:
        self.forward = partial(self.forward_dummy)

    yield

    # make everything normal
    self.__dict__.pop("forward")
    self = self.to(device_backup)


# pylint: disable=protected-access
def nncf_train_step(self, data, optimizer):
    """A helper function for train_step method of mmcv models."""
    from nncf.torch.dynamic_graph.io_handling import replicate_same_tensors

    with self._compressed_context as ctx:
        ctx.base_module_thread_local_replica = self
        _, data = replicate_same_tensors(((), data))
        if not self._in_user_dummy_forward:
            # If a user supplies own dummy forward, he is responsible for
            # correctly wrapping inputs inside it as well.
            _, data = self._strip_traced_tensors((), data)
            _, data = self._wrap_inputs_fn((), data)
        retval = self.get_nncf_wrapped_model().train_step(data, optimizer)
        retval = replicate_same_tensors(retval)
        if not self._in_user_dummy_forward:
            retval = self._wrap_outputs_fn(retval)

    # TODO: deal with kd_loss_handler in forward method of NNCFNetwork
    return retval


def no_nncf_trace_wrapper(self, fn, *args, **kwargs):  # pylint: disable=unused-argument,invalid-name
    """A wrapper function not to trace in NNCF."""

    with no_nncf_trace():
        return fn(*args, **kwargs)


def nncf_trace_wrapper(self, fn, *args, **kwargs):  # pylint: disable=unused-argument,invalid-name
    """A wrapper function to trace in NNCF."""

    with nncf_trace():
        return fn(*args, **kwargs)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# flake8: noqa
from . import (
    custom_atss_detector,
    custom_maskrcnn_detector,
    custom_single_stage_detector,
    custom_two_stage_detector,
    custom_vfnet_detector,
    custom_yolox_detector,
    l2sp_detector_mixin,
    sam_detector_mixin,
    unbiased_teacher,
)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
import torch.nn as nn
import torch.nn.functional as F
from mmdet.models import LOSSES
from mmdet.models.losses.focal_loss import py_sigmoid_focal_loss, sigmoid_focal_loss
from mmdet.models.losses.varifocal_loss import varifocal_loss


def cross_sigmoid_focal_loss(
    inputs,
    targets,
    weight=None,
    num_classes=None,
    alpha=0.25,
    gamma=2,
    reduction="mean",
    avg_factor=None,
    use_vfl=False,
    valid_label_mask=None,
):
    """
    Arguments:
       - inputs: inputs Tensor (N * C)
       - targets: targets Tensor (N), if use_vfl, then Tensor (N * C)
       - weights: weights Tensor (N), consists of (binarized label schema * weights)
       - num_classes: number of classes for training
       - alpha: focal loss alpha
       - gamma: focal loss gamma
       - reduction: default = mean
       - avg_factor: average factors
    """
    cross_mask = inputs.new_ones(inputs.shape, dtype=torch.int8)
    if valid_label_mask is not None:
        neg_mask = targets.sum(axis=1) == 0 if use_vfl else targets == num_classes
        neg_idx = neg_mask.nonzero(as_tuple=True)[0]
        cross_mask[neg_idx] = valid_label_mask[neg_idx].type(torch.int8)

    if use_vfl:
        calculate_loss_func = varifocal_loss
    else:
        if torch.cuda.is_available() and inputs.is_cuda:
            calculate_loss_func = sigmoid_focal_loss
        else:
            inputs_size = inputs.size(1)
            targets = F.one_hot(targets, num_classes=inputs_size + 1)
            targets = targets[:, :inputs_size]
            calculate_loss_func = py_sigmoid_focal_loss

    loss = (
        calculate_loss_func(inputs, targets, weight=weight, gamma=gamma, alpha=alpha, reduction="none", avg_factor=None)
        * cross_mask
    )

    if reduction == "mean":
        if avg_factor is None:
            loss = loss.mean()
        else:
            loss = loss.sum() / avg_factor
    elif reduction == "sum":
        loss = loss.sum()
    return loss


@LOSSES.register_module()
class CrossSigmoidFocalLoss(nn.Module):
    def __init__(
        self,
        use_sigmoid=True,
        num_classes=None,
        gamma=2.0,
        alpha=0.25,
        reduction="mean",
        loss_weight=1.0,
        ignore_index=None,
    ):
        super(CrossSigmoidFocalLoss, self).__init__()
        self.reduction = reduction
        self.loss_weight = loss_weight
        self.gamma = gamma
        self.alpha = alpha
        self.ignore_index = ignore_index
        self.num_classes = num_classes
        self.use_sigmoid = use_sigmoid

        self.cls_criterion = cross_sigmoid_focal_loss

    def forward(
        self,
        pred,
        targets,
        weight=None,
        reduction_override=None,
        avg_factor=None,
        use_vfl=False,
        valid_label_mask=None,
        **kwargs
    ):
        assert reduction_override in (None, "none", "mean", "sum")
        reduction = reduction_override if reduction_override else self.reduction
        loss_cls = self.loss_weight * self.cls_criterion(
            pred,
            targets,
            weight=weight,
            num_classes=self.num_classes,
            alpha=self.alpha,
            gamma=self.gamma,
            reduction=reduction,
            avg_factor=avg_factor,
            use_vfl=use_vfl,
            valid_label_mask=valid_label_mask,
        )
        return loss_cls



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
from mmcls.datasets import build_dataloader as mmcls_build_dataloader
from mmcls.datasets import build_dataset as mmcls_build_dataset

from otx.algorithms.common.adapters.mmcv.utils import (
    build_data_parallel,
    build_dataloader,
    build_dataset,
)
from otx.mpa.modules.hooks.recording_forward_hooks import (
    ActivationMapHook,
    EigenCamHook,
    ReciproCAMHook,
)
from otx.mpa.registry import STAGES
from otx.mpa.utils.logger import get_logger

from .stage import ClsStage

logger = get_logger()
EXPLAINER_HOOK_SELECTOR = {
    "eigencam": EigenCamHook,
    "activationmap": ActivationMapHook,
    "classwisesaliencymap": ReciproCAMHook,
}


@STAGES.register_module()
class ClsExplainer(ClsStage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dataset = None

    def run(self, model_cfg, model_ckpt, data_cfg, **kwargs):
        """Run explain stage
        - Configuration
        - Environment setup
        - Run explain via hooks in recording_forward_hooks
        """
        self._init_logger()
        explainer = kwargs.get("explainer")
        self.explainer_hook = EXPLAINER_HOOK_SELECTOR.get(explainer.lower(), None)
        if self.explainer_hook is None:
            raise NotImplementedError(f"Explainer algorithm {explainer} not supported!")
        logger.info(f"Explainer algorithm: {explainer}")

        cfg = self.configure(model_cfg, model_ckpt, data_cfg, training=False, **kwargs)
        logger.info("explain!")

        model_builder = kwargs.get("model_builder", None)
        outputs = self.explain(cfg, model_builder)

        return dict(outputs=outputs)

    def explain(self, cfg, model_builder=None):
        # TODO: distributed inference

        # Data loader
        self.dataset = build_dataset(cfg, "test", mmcls_build_dataset)
        explain_data_loader = build_dataloader(
            self.dataset,
            cfg,
            "test",
            mmcls_build_dataloader,
            distributed=False,
            round_up=False,
        )
        self.configure_samples_per_gpu(cfg, "test", distributed=False)

        # build the model and load checkpoint
        model = self.build_model(cfg, model_builder, fp16=cfg.get("fp16", False))
        self.extract_prob = hasattr(model, "extract_prob")
        model.eval()
        feature_model = self._get_feature_module(model)
        model = build_data_parallel(model, cfg, distributed=False)

        # InferenceProgressCallback (Time Monitor enable into Infer task)
        self.set_inference_progress_callback(model, cfg)

        eval_predictions = []
        with self.explainer_hook(feature_model) as forward_explainer_hook:
            # do inference and record intermediate fmap
            for data in explain_data_loader:
                with torch.no_grad():
                    result = model(return_loss=False, **data)
                eval_predictions.extend(result)
            saliency_maps = forward_explainer_hook.records

        assert len(eval_predictions) == len(saliency_maps), (
            "Number of elements should be the same, however, number of outputs are "
            f"{len(eval_predictions)}, and {len(saliency_maps)}"
        )

        outputs = dict(
            eval_predictions=eval_predictions,
            saliency_maps=saliency_maps,
        )
        return outputs



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import otx.mpa.modules.datasets.pipelines.transforms.augmix
import otx.mpa.modules.datasets.pipelines.transforms.ote_transforms
import otx.mpa.modules.datasets.pipelines.transforms.random_augment
import otx.mpa.modules.datasets.pipelines.transforms.twocrop_transform
import otx.mpa.modules.hooks
import otx.mpa.modules.models.classifiers
import otx.mpa.modules.models.heads.custom_cls_head
import otx.mpa.modules.models.heads.custom_hierarchical_linear_cls_head
import otx.mpa.modules.models.heads.custom_hierarchical_non_linear_cls_head
import otx.mpa.modules.models.heads.custom_multi_label_linear_cls_head
import otx.mpa.modules.models.heads.custom_multi_label_non_linear_cls_head
import otx.mpa.modules.models.heads.non_linear_cls_head
import otx.mpa.modules.models.heads.semisl_cls_head
import otx.mpa.modules.models.heads.supcon_cls_head
import otx.mpa.modules.models.losses.asymmetric_angular_loss_with_ignore
import otx.mpa.modules.models.losses.asymmetric_loss_with_ignore
import otx.mpa.modules.models.losses.barlowtwins_loss
import otx.mpa.modules.models.losses.cross_entropy_loss
import otx.mpa.modules.models.losses.ib_loss
import otx.mpa.modules.optimizer.lars

# flake8: noqa
from . import (
    evaluator,
    explainer,
    exporter,
    incremental,
    inferrer,
    semisl,
    stage,
    trainer,
)



"""This module contains functions for basic operations."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from typing import Dict, List, Optional, Tuple

import numpy as np

from otx.api.entities.label import LabelEntity
from otx.api.entities.shapes.rectangle import Rectangle

#: Dictionary storing a number for each label. The ``None`` key represents "all labels"
NumberPerLabel = Dict[Optional[LabelEntity], int]


def get_intersections_and_cardinalities(
    references: List[np.ndarray],
    predictions: List[np.ndarray],
    labels: List[LabelEntity],
) -> Tuple[NumberPerLabel, NumberPerLabel]:
    """Returns all intersections and cardinalities between reference masks and prediction masks.

    Intersections and cardinalities are each returned in a dictionary mapping each label to its corresponding
    number of intersection/cardinality pixels

    Args:
        references (List[np.ndarray]): reference masks,s one mask per image
        predictions (List[np.ndarray]): prediction masks, one mask per image
        labels (List[LabelEntity]): labels in input masks

    Returns:
        Tuple[NumberPerLabel, NumberPerLabel]: (all_intersections, all_cardinalities)
    """
    all_intersections: NumberPerLabel = {label: 0 for label in labels}
    all_intersections[None] = 0
    all_cardinalities: NumberPerLabel = {label: 0 for label in labels}
    all_cardinalities[None] = 0
    for reference, prediction in zip(references, predictions):
        intersection = np.where(reference == prediction, reference, 0)
        all_intersections[None] += np.count_nonzero(intersection)
        all_cardinalities[None] += np.count_nonzero(reference) + np.count_nonzero(prediction)
        for i, label in enumerate(labels):
            label_num = i + 1
            all_intersections[label] += np.count_nonzero(intersection == label_num)
            reference_area = np.count_nonzero(reference == label_num)
            prediction_area = np.count_nonzero(prediction == label_num)
            all_cardinalities[label] += reference_area + prediction_area
    return all_intersections, all_cardinalities


def intersection_box(box1: Rectangle, box2: Rectangle) -> Optional[List[float]]:
    """Calculate the intersection box of two bounding boxes.

    Args:
        box1: a Rectangle that represents the first bounding box
        box2: a Rectangle that represents the second bounding box

    Returns:
        a Rectangle that represents the intersection box if inputs have
        a valid intersection, else None
    """
    x_left = max(box1.x1, box2.x1)
    y_top = max(box1.y1, box2.y1)
    x_right = min(box1.x2, box2.x2)
    y_bottom = min(box1.y2, box2.y2)
    if x_right <= x_left or y_bottom <= y_top:
        return None
    return [x_left, y_top, x_right, y_bottom]


def intersection_over_union(box1: Rectangle, box2: Rectangle, intersection: Optional[List[float]] = None) -> float:
    """Calculate the Intersection over Union (IoU) of two bounding boxes.

    Args:
        box1: a Rectangle representing a bounding box
        box2: a Rectangle representing a second bounding box
        intersection: precomputed intersection between two boxes (see
            intersection_box function), if exists.

    Returns:
        intersection-over-union of box1 and box2
    """
    iou = 0.0
    if intersection is None:
        intersection = intersection_box(box1, box2)
    if intersection is not None:
        intersection_area = (intersection[2] - intersection[0]) * (intersection[3] - intersection[1])
        box1_area = (box1.x2 - box1.x1) * (box1.y2 - box1.y1)
        box2_area = (box2.x2 - box2.x1) * (box2.y2 - box2.y1)
        union_area = float(box1_area + box2_area - intersection_area)
        if union_area != 0:
            iou = intersection_area / union_area
    if iou < 0.0 or iou > 1.0:
        raise ValueError(f"intersection over union should be in range [0,1], instead got iou={iou}")
    return iou


def precision_per_class(matrix: np.ndarray) -> np.ndarray:
    """Compute the precision per class based on the confusion matrix.

    Args:
        matrix: the computed confusion matrix

    Returns:
        the precision (per class), defined as TP/(TP+FP)
    """
    if not matrix.shape[0] == matrix.shape[1]:
        # If the matrix is not square (there is a column for "other" label), the "other" column is deleted.
        # Otherwise, there will be 3 elements in TP and 4 in TP+FP meaning they can't be divided.
        matrix = np.delete(matrix, -1, 1)

    tp_per_class = matrix.diagonal()
    sum_tp_fp_per_class = matrix.sum(0)
    return divide_arrays_with_possible_zeros(tp_per_class, sum_tp_fp_per_class)


def recall_per_class(matrix: np.ndarray) -> np.ndarray:
    """Compute the recall per class based on the confusion matrix.

    Args:
        matrix: the computed confusion matrix

    Returns:
        the recall (per class), defined as TP/(TP+FN)
    """
    tp_per_class = matrix.diagonal()
    sum_tp_fn_per_class = matrix.sum(1)
    return divide_arrays_with_possible_zeros(tp_per_class, sum_tp_fn_per_class)


def divide_arrays_with_possible_zeros(array1: np.ndarray, array2: np.ndarray) -> np.ndarray:
    """Sometimes the denominator in the precision or recall computation can contain a zero.

    In that case, a zero is returned for that element (https://stackoverflow.com/a/32106804).

    Args:
        array1: the numerator
        array2: the denominator

    Returns:
        the divided arrays (numerator/denominator) with a value of zero
        where the denominator was zero.
    """
    with np.errstate(divide="ignore", invalid="ignore"):
        result = np.true_divide(array1, array2)
        result[result == np.inf] = 0  # If the denominator is a float, np.inf is returned
        result = np.nan_to_num(result)  # If the denominator is an int, np.nan is returned
    return result



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Any, Dict, Optional


class Registry:
    REGISTERED_NAME_ATTR = "_registered_name"

    def __init__(self, name, add_name_as_attr=False):
        self._name = name
        self._registry_dict = {}
        self._add_name_as_attr = add_name_as_attr

    @property
    def registry_dict(self) -> Dict[Any, Any]:
        return self._registry_dict

    def _register(self, obj: Any, name: Any):
        if name in self._registry_dict:
            raise KeyError("{} is already registered in {}".format(name, self._name))
        self._registry_dict[name] = obj

    def register(self, name: Optional[Any] = None):
        def wrap(obj):
            cls_name = name
            if cls_name is None:
                cls_name = obj.__name__
            if self._add_name_as_attr:
                setattr(obj, self.REGISTERED_NAME_ATTR, cls_name)
            self._register(obj, cls_name)
            return obj

        return wrap

    def get(self, key: Any) -> Any:
        if key not in self._registry_dict:
            self._key_not_found(key)
        return self._registry_dict[key]

    def _key_not_found(self, key: Any):
        raise KeyError("{} is not found in {}".format(key, self._name))

    def __contains__(self, item):
        return item in self._registry_dict.values()



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from ..builder import SCALAR_SCHEDULERS
from .base import BaseScalarScheduler


@SCALAR_SCHEDULERS.register_module()
class ConstantScalarScheduler(BaseScalarScheduler):
    """The learning rate remains constant over time.

    The learning rate equals the scale.
    Args:
        scale (float): The learning rate scale.
    """

    def __init__(self, scale: float = 30.0):
        super(ConstantScalarScheduler, self).__init__()

        self._end_s = scale
        assert self._end_s > 0.0

    def _get_value(self, step, epoch_size):
        return self._end_s



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest

from otx.algorithms.common.utils import get_task_class, load_template
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestCommonUtilsInputParamsValidation:
    @e2e_pytest_unit
    def test_get_task_class_input_params_validation(self):
        """
        <b>Description:</b>
        Check "get_task_class" function input parameters validation

        <b>Input data:</b>
        "path" non string-type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "get_task_class" function
        """
        with pytest.raises(ValueError):
            get_task_class(path=1)  # type: ignore

    @e2e_pytest_unit
    def test_load_template_params_validation(self):
        """
        <b>Description:</b>
        Check "load_template" function input parameters validation

        <b>Input data:</b>
        "path" unexpected string with yaml file object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "load_template" function
        """
        for incorrect_parameter in [
            # Unexpected integer is specified as "path" parameter
            1,
            # Empty string is specified as "path" parameter
            "",
            # Path to non-existing file is specified as "path" parameter
            "./non_existing.yaml",
            # Path to non-yaml file is specified as "path" parameter
            "./unexpected_type.jpg",
            # Path Null character is specified in "path" parameter
            "./null\0char.yaml",
            # Path with non-printable character is specified as "path" parameter
            "./non\nprintable.yaml",
        ]:
            with pytest.raises(ValueError):
                load_template(incorrect_parameter)
